//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Coord/ICoordSystem.cpp
//! @brief     Implements ICoordSystem class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Coord/ICoordSystem.h"
#include "Base/Util/Assert.h"
#include "Device/Coord/AxisNames.h"
#include "Device/Data/Datafield.h"
#include <sstream>

ICoordSystem::~ICoordSystem() = default;

std::vector<IAxis*> ICoordSystem::defaultAxes() const
{
    return convertedAxes(defaultUnits());
}

std::vector<IAxis*> ICoordSystem::convertedAxes(Coords units) const
{
    std::vector<IAxis*> result;
    for (size_t i = 0; i < rank(); ++i)
        result.emplace_back(createConvertedAxis(i, units));
    return result;
}

std::string ICoordSystem::axisName(size_t i_axis, const Coords units) const
{
    const auto& name_maps = createNameMaps();
    ASSERT(i_axis < name_maps.size());
    const auto& name_map = name_maps[i_axis];
    const auto& it = name_map.find(units == Coords::UNDEFINED ? defaultUnits() : units);
    ASSERT(it != name_map.cend());
    return it->second;
}

Datafield* ICoordSystem::createConvertedData(const Datafield& data, Coords units) const
{
    return new Datafield(convertedAxes(units), data.flatVector());
}

void ICoordSystem::throwUnitsError(std::string method, std::vector<Coords> available) const
{
    std::stringstream ss;
    ss << "Unit type error in " << method
       << ": unknown or unsupported unit type. Available units "
          "are:\n";
    for (auto unit : available)
        ss << axisUnitLabel.at(unit) << "\n";
    throw std::runtime_error(ss.str());
}

Coords ICoordSystem::substituteDefaultUnits(Coords units) const
{
    return units == Coords::UNDEFINED ? defaultUnits() : units;
}
