//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/DataItem.cpp
//! @brief     Implements class IntensityDataItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/DataItem.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/Util/Error.h"

const QString x_axis_default_name = "X [nbins]";
const QString y_axis_default_name = "Y [nbins]";

DataItem::DataItem(const QString& modelType)
    : SessionItem(modelType)
{
    // name of the file used to serialize given IntensityDataItem
    addProperty(P_FILE_NAME, "undefined");
    addProperty(P_AXES_UNITS, ComboProperty::fromList({"nbins"}).variant());
    addProperty<BasicAxisItem>(P_XAXIS);
    addProperty<AmplitudeAxisItem>(P_YAXIS);

    setXaxisTitle(x_axis_default_name);
    setYaxisTitle(y_axis_default_name);
}

void DataItem::setDatafield(Datafield* data)
{
    std::unique_lock<std::mutex> lock(m_update_data_mutex);
    m_datafield.reset(data);

    setLastModified(QDateTime::currentDateTime());
    emit datafieldChanged();
}

void DataItem::setRawDataVector(const std::vector<double>& data)
{
    if (m_datafield->size() != data.size())
        throw Error("DataItem::setVector() -> Error. "
                    "Different data size.");
    std::unique_lock<std::mutex> lock(m_update_data_mutex);
    m_datafield->setVector(data);

    setLastModified(QDateTime::currentDateTime());
    emit rawDataVectorChanged(data);
}

QString DataItem::fileName() const
{
    return getItemValue(P_FILE_NAME).toString();
}

void DataItem::setFileName(const QString& filename)
{
    setItemValue(P_FILE_NAME, filename);

    setLastModified(QDateTime::currentDateTime());
    emit fileNameChanged(filename);
}

QDateTime DataItem::lastModified() const
{
    return m_last_modified;
}

bool DataItem::containsNonXMLData() const
{
    return static_cast<bool>(m_datafield);
}

bool DataItem::loadData(const QString& projectDir)
{
    const QString filename = fileName(projectDir);
    auto* data = IOFactory::readDatafield(filename.toStdString());
    if (!data)
        return false;
    setDatafield(data);
    return true;
}

bool DataItem::saveData(const QString& projectDir)
{
    if (!containsNonXMLData())
        return false;

    std::unique_lock<std::mutex> lock(m_update_data_mutex);
    std::unique_ptr<Datafield> clone(datafield()->clone());
    lock.unlock();
    IOFactory::writeDatafield(*clone, fileName(projectDir).toStdString());
    return true;
}

void DataItem::setLastModified(const QDateTime& dtime)
{
    m_last_modified = dtime;
}

int DataItem::xSize() const
{
    return xAxisItem()->binCount();
}

int DataItem::ySize() const
{
    return yAxisItem()->binCount();
}

double DataItem::lowerX() const
{
    return xAxisItem()->min();
}

double DataItem::upperX() const
{
    return xAxisItem()->max();
}

void DataItem::setLowerX(double value)
{
    xAxisItem()->setMin(value);
    emit itemAxesRangeChanged();
}

void DataItem::setUpperX(double value)
{
    xAxisItem()->setMax(value);
    emit itemAxesRangeChanged();
}

double DataItem::lowerY() const
{
    return yAxisItem()->min();
}

double DataItem::upperY() const
{
    return yAxisItem()->max();
}

void DataItem::setLowerY(double value)
{
    yAxisItem()->setMin(value);
    emit itemAxesRangeChanged();
}

void DataItem::setUpperY(double value)
{
    yAxisItem()->setMax(value);
    emit itemAxesRangeChanged();
}

void DataItem::copyXRangeFromItem(DataItem* sourceItem)
{
    if (sourceItem == this)
        return;
    setLowerX(sourceItem->lowerX());
    setUpperX(sourceItem->upperX());
}

void DataItem::copyYRangeFromItem(DataItem* sourceItem)
{
    if (sourceItem == this)
        return;
    setLowerY(sourceItem->lowerY());
    setUpperY(sourceItem->upperY());
}

void DataItem::copyXYRangesFromItem(DataItem* sourceItem)
{
    copyXRangeFromItem(sourceItem);
    copyYRangeFromItem(sourceItem);
}

QString DataItem::currentAxesUnits() const
{
    return axesUnitsCombo().currentValue();
}

void DataItem::setCurrentAxesUnits(const QString& units)
{
    ComboProperty combo = axesUnitsCombo();
    combo.setCurrentValue(units);
    setAxesUnitsCombo(combo);
}

ComboProperty DataItem::axesUnitsCombo() const
{
    return getItemValue(P_AXES_UNITS).value<ComboProperty>();
}

void DataItem::setAxesUnitsCombo(const ComboProperty& unitsCombo)
{
    setItemValue(P_AXES_UNITS, unitsCombo.variant());
    emit axesUnitsChanged(this);
}

const BasicAxisItem* DataItem::xAxisItem() const
{
    return item<BasicAxisItem>(P_XAXIS);
}

BasicAxisItem* DataItem::xAxisItem()
{
    return item<BasicAxisItem>(P_XAXIS);
}

const AmplitudeAxisItem* DataItem::yAxisItem() const
{
    return item<AmplitudeAxisItem>(P_YAXIS);
}

AmplitudeAxisItem* DataItem::yAxisItem()
{
    return item<AmplitudeAxisItem>(P_YAXIS);
}

QString DataItem::XaxisTitle() const
{
    return xAxisItem()->title();
}

QString DataItem::YaxisTitle() const
{
    return yAxisItem()->title();
}

void DataItem::setXaxisTitle(const QString& title)
{
    xAxisItem()->setTitle(title);
}

void DataItem::setYaxisTitle(const QString& title)
{
    yAxisItem()->setTitle(title);
}

void DataItem::setAxesRangeToData()
{
    setLowerX(xMin());
    setUpperX(xMax());
    setLowerY(yMin());
    setUpperY(yMax());
}
