//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/SpecularDataItem.cpp
//! @brief     Implements class SpecularDataItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/SpecularDataItem.h"
#include "Base/Axis/IAxis.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Support/IO/ImportDataInfo.h"
#include "GUI/Util/Error.h"

namespace {

// set of simple colors for representation of 1D graphs
const QMap<QString, Qt::GlobalColor> color_map = {
    {"Blue", Qt::GlobalColor::blue},     {"Black", Qt::GlobalColor::black},
    {"Red", Qt::GlobalColor::darkRed},   {"Cyan", Qt::GlobalColor::darkCyan},
    {"Gray", Qt::GlobalColor::darkGray}, {"Magenta", Qt::GlobalColor::darkMagenta}};
// scatters for representation of 1D graphs
const QMap<QString, QCPScatterStyle::ScatterShape> scatter_map = {
    {"None", QCPScatterStyle::ScatterShape::ssNone},
    {"Disc", QCPScatterStyle::ScatterShape::ssDisc},
    {"Circle", QCPScatterStyle::ScatterShape::ssCircle},
    {"Cross", QCPScatterStyle::ScatterShape::ssCross},
    {"Diamond", QCPScatterStyle::ScatterShape::ssDiamond},
    {"Star", QCPScatterStyle::ScatterShape::ssStar}};
// connection lines for representation of 1D graphs
const QMap<QString, QCPGraph::LineStyle> line_map = {
    {"None", QCPGraph::LineStyle::lsNone},
    {"Line", QCPGraph::LineStyle::lsLine},
    {"StepLeft", QCPGraph::LineStyle::lsStepLeft},
    {"StepRight", QCPGraph::LineStyle::lsStepRight},
    {"StepCenter", QCPGraph::LineStyle::lsStepCenter},
    {"Impulse", QCPGraph::LineStyle::lsImpulse}};

} // namespace

SpecularDataItem::SpecularDataItem(const QString& modelType)
    : DataItem(modelType)
{
    addProperty(P_TITLE, QString());

    addProperty(P_LINESTYLE, line_map.key(QCPGraph::LineStyle::lsLine));
    addProperty(P_COLOR, color_map.key(Qt::GlobalColor::blue));
    addProperty(P_THICKNESS, 1.5);
    addProperty(P_SCATTER, scatter_map.key(QCPScatterStyle::ScatterShape::ssNone));
    addProperty(P_SCATTERSIZE, 5.0);

    setYaxisTitle("Signal [a.u]");
    setSimuPlotStyle();
}

SpecularDataItem::SpecularDataItem()
    : SpecularDataItem(M_TYPE)
{
}

void SpecularDataItem::setDatafield(Datafield* data)
{
    if (data != nullptr) {
        if (data->rank() != 1)
            throw Error("Error in SpecularDataItem::setDatafield: cannot handle non-1D data");
        DataItem::setDatafield(data);
        updateAxesZoomLevel();
    } else
        DataItem::setDatafield(data);
}

double SpecularDataItem::xMin() const
{
    const double defaultXmin(0.0);
    return m_datafield ? m_datafield->axis(0).min() : defaultXmin;
}

double SpecularDataItem::xMax() const
{
    const double defaultXmax(1.0);
    return m_datafield ? m_datafield->axis(0).max() : defaultXmax;
}

double SpecularDataItem::yMin() const
{
    return dataRange().first;
}

double SpecularDataItem::yMax() const
{
    return dataRange().second;
}

bool SpecularDataItem::isLog() const
{
    return yAxisItem()->isLogScale();
}

void SpecularDataItem::setLog(bool islog)
{
    yAxisItem()->setLogScale(islog);
}

void SpecularDataItem::updateCoords(const InstrumentItem* instrument)
{
    GUI::Model::JobItemUtils::updateDataAxes(this, instrument);
}

std::vector<int> SpecularDataItem::shape() const
{
    return {xSize()};
}

void SpecularDataItem::reset(ImportDataInfo data)
{
    ComboProperty combo = ComboProperty() << data.unitsLabel();
    setAxesUnitsCombo(combo);
    setXaxisTitle(data.axisLabel(0));
    setYaxisTitle(data.axisLabel(1));
    setDatafield(std::move(data).intensityData().release());
    setAxesRangeToData();
}

QCPGraph::LineStyle SpecularDataItem::lineStyle()
{
    const QString& lineType = getItemValue(P_LINESTYLE).toString();
    ASSERT(line_map.contains(lineType));
    return QCPGraph::LineStyle(line_map.value(lineType));
}

void SpecularDataItem::setLineStyle(QCPGraph::LineStyle lineStyle)
{
    ASSERT(line_map.values().contains(lineStyle));
    setItemValue(P_LINESTYLE, line_map.key(lineStyle));
}

QColor SpecularDataItem::color()
{
    const QString& colorName = getItemValue(P_COLOR).toString();
    ASSERT(color_map.contains(colorName));
    return QColor(color_map.value(colorName));
}

void SpecularDataItem::setColor(Qt::GlobalColor color)
{
    ASSERT(color_map.values().contains(color));
    setItemValue(P_COLOR, color_map.key(color));
}

double SpecularDataItem::thickness()
{
    return getItemValue(P_THICKNESS).toDouble();
}

void SpecularDataItem::setThickness(double thickness)
{
    setItemValue(P_THICKNESS, thickness);
}

QCPScatterStyle::ScatterShape SpecularDataItem::scatter()
{
    const QString& scatterName = getItemValue(P_SCATTER).toString();
    ASSERT(scatter_map.contains(scatterName));
    return QCPScatterStyle::ScatterShape(scatter_map.value(scatterName));
}

void SpecularDataItem::setScatter(QCPScatterStyle::ScatterShape scatter)
{
    ASSERT(scatter_map.values().contains(scatter));
    setItemValue(P_SCATTER, scatter_map.key(scatter));
}

double SpecularDataItem::scatterSize()
{
    return getItemValue(P_SCATTERSIZE).toDouble();
}

void SpecularDataItem::setScatterSize(double scatterSize)
{
    setItemValue(P_SCATTERSIZE, scatterSize);
}

void SpecularDataItem::setSimuPlotStyle()
{
    setScatter(QCPScatterStyle::ScatterShape::ssNone);
    setColor(Qt::GlobalColor::blue);
    setLineStyle(QCPGraph::LineStyle::lsLine);
}

void SpecularDataItem::setDiffPlotStyle()
{
    setScatter(QCPScatterStyle::ScatterShape::ssNone);
    setColor(Qt::GlobalColor::black);
    setLineStyle(QCPGraph::LineStyle::lsLine);

    setYaxisTitle("Relative difference");
}

void SpecularDataItem::setRealPlotStyle()
{
    setScatter(QCPScatterStyle::ScatterShape::ssDisc);
    setColor(Qt::GlobalColor::black);
    setLineStyle(QCPGraph::LineStyle::lsNone);
}

//! Sets zoom range of X,Y axes, if it was not yet defined.

void SpecularDataItem::updateAxesZoomLevel()
{
    // set zoom range of x-axis to min, max values if it was not set already
    if (upperX() < lowerX()) {
        setLowerX(xMin());
        setUpperX(xMax());
    }

    // set zoom range of y-axis to min, max values if it was not set already
    if (upperY() < lowerY()) {
        setLowerY(yMin());
        setUpperY(yMax());
    }

    const int nx = static_cast<int>(m_datafield->axis(0).size());
    xAxisItem()->setBinCount(nx);
}

QPair<double, double> SpecularDataItem::dataRange() const
{
    const double default_min = 0.0;
    const double default_max = 1.0;
    const Datafield* data = datafield();
    if (!data)
        return QPair<double, double>(default_min, default_max);

    const auto vec = data->flatVector();
    double min(*std::min_element(vec.cbegin(), vec.cend()));
    double max(*std::max_element(vec.cbegin(), vec.cend()));

    min /= 2.0;
    min = std::numeric_limits<double>::epsilon() < min ? min : default_min;
    max *= 2.0;
    max = max > min ? max : default_max;

    return QPair<double, double>(min, max);
}

void SpecularDataItem::resetView()
{
    setAxesRangeToData();
}
