//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/DistributionItems.h
//! @brief     Defines class DistributionItem and several subclasses
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_DISTRIBUTIONITEMS_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_DISTRIBUTIONITEMS_H

#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Model/Descriptor/UIntProperty.h"
#include <QVector>
#include <memory>

class IDistribution1D;
class IRangedDistribution;
class Streamer;

class DistributionItem {
public:
    virtual ~DistributionItem() = default;
    DistributionItem();

    template <typename T>
    bool is() const
    {
        return dynamic_cast<const T*>(this) != nullptr;
    }

    virtual std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const = 0;

    virtual void initDistribution(double) {}

    virtual void setLimits(const RealLimits& limits);
    RealLimits limits() const;

    //! Set the unit of the distributed value
    virtual void setUnit(const std::variant<QString, Unit>& unit) = 0;

    UINT_PROPERTY(numberOfSamples, NumberOfSamples);
    DOUBLE_PROPERTY(sigmaFactor, SigmaFactor);
    bool hasSigmaFactor() const;

    //! Serialization of contents.
    //!
    //! Important: limits and a changed unit (with setUnit) will not be serialized here. They have
    //! to be set again by the owner of DistributionItem after reading it
    void serialize(Streamer& s);

    virtual DoubleDescriptors distributionValues(bool withMean = true) const = 0;

protected:
    void initSigmaFactor();
    void addToSerialization(DoubleProperty* d);

    RealLimits m_limits;

private:
    QVector<DoubleProperty*> m_serializationProperties; //! properties to be serialized
};

class SymmetricResolutionItem : public DistributionItem {
public:
    SymmetricResolutionItem(double mean, int decimals = 3, const QString& meanLabel = "Mean");
    void setUnit(const std::variant<QString, Unit>& unit) override;
    void setLimits(const RealLimits& limits) override;

    DOUBLE_PROPERTY(mean, Mean);
    void setMeanDecimals(uint d);

    virtual std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const = 0;
    virtual double deviation(double scale) const = 0;
};

class DistributionNoneItem : public SymmetricResolutionItem {
public:
    DistributionNoneItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;
    void initDistribution(double value) override;

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

class DistributionGateItem : public DistributionItem {
public:
    DistributionGateItem();

    void setUnit(const std::variant<QString, Unit>& unit) override;

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    void initDistribution(double value) override;

    DOUBLE_PROPERTY(minimum, Minimum);
    DOUBLE_PROPERTY(maximum, Maximum);
    void setRange(double min, double max);

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

class DistributionLorentzItem : public SymmetricResolutionItem {
public:
    DistributionLorentzItem();
    void setUnit(const std::variant<QString, Unit>& unit) override;

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;

    DOUBLE_PROPERTY(hwhm, Hwhm);

    void initDistribution(double value) override;

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

class DistributionGaussianItem : public SymmetricResolutionItem {
public:
    DistributionGaussianItem();
    void setUnit(const std::variant<QString, Unit>& unit) override;

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;

    DOUBLE_PROPERTY(standardDeviation, StandardDeviation);

    void initDistribution(double value) override;

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

class DistributionLogNormalItem : public DistributionItem {
public:
    DistributionLogNormalItem();
    void setUnit(const std::variant<QString, Unit>& unit) override;

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    void initDistribution(double value) override;

    DOUBLE_PROPERTY(median, Median);
    DOUBLE_PROPERTY(scaleParameter, ScaleParameter);

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

class DistributionCosineItem : public SymmetricResolutionItem {
public:
    DistributionCosineItem();
    void setUnit(const std::variant<QString, Unit>& unit) override;

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;
    void initDistribution(double value) override;

    DOUBLE_PROPERTY(sigma, Sigma);

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

class DistributionTrapezoidItem : public DistributionItem {
public:
    DistributionTrapezoidItem();
    void setUnit(const std::variant<QString, Unit>& unit) override;

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    void initDistribution(double value) override;

    DOUBLE_PROPERTY(center, Center);
    DOUBLE_PROPERTY(leftWidth, LeftWidth);
    DOUBLE_PROPERTY(middleWidth, MiddleWidth);
    DOUBLE_PROPERTY(rightWidth, RightWidth);

    DoubleDescriptors distributionValues(bool withMean = true) const override;
};

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_DISTRIBUTIONITEMS_H
