//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/UIntDescriptor.h
//! @brief     Defines class UIntDescriptor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_UINTDESCRIPTOR_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_UINTDESCRIPTOR_H

#include "Fit/Param/RealLimits.h"
#include "GUI/Support/Type/Unit.h"
#include <QString>
#include <functional>
#include <variant>

class SessionItem;

//! Describes properties of a uint value which are necessary to allow GUI representation, editing
//! the value, undo/redo, unit conversion.
//!
//! By using this class, the underlying data scheme is hidden from the user of the data. This e.g.
//! eases SessionItem migration. The underlying implementation can be a SessionItem, a simple uint
//! member, or any other construction to hold a uint value.
class UIntDescriptor {
public:
    UIntDescriptor() = default;
    //! Operates on a session item. The settings (like limits) are taken from the session
    //! item.
    //! Only for easier migration. Should be removed after SessionItem refactoring.
    UIntDescriptor(SessionItem* item, const std::variant<QString, Unit>& unit);

    //! Operates on a session item. The settings (like limits) are taken from the session
    //! item.
    //! Only for easier migration. Should be removed after SessionItem refactoring.
    UIntDescriptor(const QString& label, SessionItem* item,
                   const std::variant<QString, Unit>& unit);

    UIntDescriptor(const QString& label, const uint* var, const std::variant<QString, Unit>& unit);

private:
    //! Operates on any kind of storage (e.g. session items), by using setter/getter methods
    UIntDescriptor(QString label, QString tooltip, const RealLimits& limits,
                   std::function<void(uint)> setter, std::function<uint()> getter,
                   const std::variant<QString, Unit>& unit);

    //! Operates on any kind of storage (e.g. session items), by using setter/getter methods
    //! limits is set to nonnegative
    UIntDescriptor(const QString& label, const QString& tooltip, std::function<void(uint)> setter,
                   std::function<uint()> getter, const std::variant<QString, Unit>& unit);

public:
    //! Return the current value of the handled parameter.
    operator uint() const;

    QString label;                                     //!< A label text (short, no trailing colon)
    QString tooltip;                                   //!< Tooltip text
    RealLimits limits;                                 //!< Limits of the value.
    std::function<void(uint)> set = nullptr;           //!< function to set the value
    std::function<uint()> get = nullptr;               //!< function to get the current value
    std::variant<QString, Unit> unit = Unit::unitless; //!< Unit of the value (internal unit only!)
    std::function<QString()> path =
        nullptr; //!< Path describing  this value. Used e.g. for undo/redo
};

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_UINTDESCRIPTOR_H
