//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/JobItem.cpp
//! @brief     Implements class JobItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/JobItem.h"
#include "Base/Util/Assert.h"
#include "Device/Detector/SimulationAreaIterator.h"
#include "GUI/Model/CatDevice/InstrumentItemCatalog.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/Instrument.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Model/SessionXML.h" // writeItemAndChildItems
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Support/IO/ItemFileNameUtils.h"
#include "GUI/Support/XML/Streamer.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/DeserializationException.h"
#include "GUI/Util/Error.h"

#include <stdexcept> // domain_error

namespace {
namespace Tags {

const QString SimulationOptions("SimulationOptions");
const QString Materials("Materials");
const QString ParameterContainer("ParameterContainer");
const QString Sample("Sample");
const QString Instrument("Instrument");
const QString value("value");
const QString Identifier("Id");
const QString Comments("Comments");
const QString PresentationType("PresentationType");
const QString Activity("Activity");
const QString Duration("Duration");
const QString BeginTime("Begin");
const QString EndTime("End");
const QString Status("Status");
const QString Progress("Progress");
const QString SimulatedData("SimulatedData");
const QString RealItem("RealItem");
const QString DiffData("DiffData"); // diff. between simulation result and real data
const QString FitSuite("FitSuite");
} // namespace Tags
} // namespace

JobItem::JobItem()
    : JobRealBase(M_TYPE)
{
    setItemName(M_TYPE);
    REFACTORED = true;
}

QString JobItem::getIdentifier() const
{
    return m_identifier;
}

void JobItem::setIdentifier(const QString& identifier)
{
    m_identifier = identifier;
}

QString JobItem::jobName() const
{
    return itemName();
}

void JobItem::setJobName(const QString& name)
{
    setItemName(name);
    updateDataFileName();
    emit jobNameChanged(name);
}

JobStatus JobItem::getStatus() const
{
    return m_status;
}

void JobItem::setStatus(JobStatus status)
{
    m_status = status;
    if (status == JobStatus::Failed) {
        if (DataItem* dataItem = simulatedDataItem()) {
            if (dataItem->datafield())
                dataItem->datafield()->setAllTo(0.0);
            emit dataItem->emitDataChanged();
        }
    }
    emit jobStatusChanged(status);
}

bool JobItem::isIdle() const
{
    return getStatus() == JobStatus::Idle;
}

bool JobItem::isRunning() const
{
    return getStatus() == JobStatus::Running;
}

bool JobItem::isCompleted() const
{
    return getStatus() == JobStatus::Completed;
}

bool JobItem::isCanceled() const
{
    return getStatus() == JobStatus::Canceled;
}

bool JobItem::isFailed() const
{
    return getStatus() == JobStatus::Failed;
}

bool JobItem::isFitting() const
{
    return getStatus() == JobStatus::Fitting;
}

bool JobItem::isValidForFitting()
{
    return bool(m_realItem);
}

QDateTime JobItem::beginTime() const
{
    return m_beginTime;
}

void JobItem::setBeginTime(const QDateTime& begin_time)
{
    m_beginTime = begin_time;
    emit jobBeginTimeChanged(begin_time);
}

QDateTime JobItem::endTime() const
{
    return m_endTime;
}

void JobItem::setEndTime(const QDateTime& end_time)
{
    m_endTime = end_time;
    emit jobEndTimeChanged(end_time);
}

std::optional<size_t> JobItem::duration() const
{
    QDateTime begin_time = beginTime();
    QDateTime end_time = endTime();
    if (begin_time.isValid() && end_time.isValid() && begin_time < end_time)
        return begin_time.msecsTo(end_time);
    return std::nullopt;
}

QString JobItem::getComments() const
{
    return m_comments;
}

void JobItem::setComments(const QString& comments)
{
    m_comments = comments;
    emit jobCommentsChanged(comments);
}

int JobItem::getProgress() const
{
    return m_progress;
}

void JobItem::setProgress(int progress)
{
    m_progress = progress;
    emit jobProgressChanged(progress);
}

bool JobItem::runImmediately() const
{
    return simulationOptionsItem().runImmediately();
}

bool JobItem::runInBackground() const
{
    return simulationOptionsItem().runInBackground();
}

QString JobItem::presentationType() const
{
    return m_presentationType;
}

void JobItem::setPresentationType(const QString& type)
{
    m_presentationType = type;
}

QString JobItem::activity() const
{
    return m_activity;
}

void JobItem::setActivity(const QString& activity)
{
    m_activity = activity;
}

bool JobItem::isSpecularJob() const
{
    return instrumentItem()->is<SpecularInstrumentItem>();
}

bool JobItem::isIntensityJob() const
{
    return instrumentItem()->is<GISASInstrumentItem>()
           || instrumentItem()->is<OffspecInstrumentItem>()
           || instrumentItem()->is<DepthProbeInstrumentItem>();
}

MultiLayerItem* JobItem::sampleItem()
{
    return &m_sampleItem;
}

void JobItem::copySampleIntoJob(const MultiLayerItem* sample)
{
    m_sampleItem.initFrom(sample);
}

InstrumentItem* JobItem::instrumentItem() const
{
    return m_instrument.get();
}

Instrument2DItem* JobItem::instrument2DItem()
{
    auto* result = dynamic_cast<Instrument2DItem*>(instrumentItem());
    ASSERT(result);
    return result;
}

void JobItem::copyInstrumentIntoJob(const InstrumentItem* instrument)
{
    m_instrument.reset(instrument->createCopy());
    m_instrument.get()->setId(QUuid::createUuid().toString());
}

const SimulationOptionsItem& JobItem::simulationOptionsItem() const
{
    return m_simulationOptionsItem;
}

void JobItem::copySimulationOptionsIntoJob(const SimulationOptionsItem& options)
{
    m_simulationOptionsItem = options;
}

void JobItem::setResults(const SimulationResult& result)
{
    GUI::Model::JobItemUtils::setResults(simulatedDataItem(), result);
    m_simulatedDataItem->setParentAndModel(this, model());
    updateDataFileName();
}

FitSuiteItem* JobItem::fitSuiteItem()
{
    return m_fitSuiteItem.get();
}

FitSuiteItem* JobItem::createFitSuiteItem()
{
    if (m_fitSuiteItem)
        throw Error("JobItem::createFitSuiteItem() -> Error. Attempt to create "
                    "a second FitSuiteItem.");

    m_fitSuiteItem = std::make_unique<FitSuiteItem>();
    m_fitSuiteItem->setParentAndModel(this, model());
    return m_fitSuiteItem.get();
}

void JobItem::createFitContainers()
{
    FitSuiteItem* fitSuiteItem = createFitSuiteItem();

    fitSuiteItem->createFitParametersContainer();
    fitSuiteItem->createMinimizerContainer();
}

ParameterContainerItem* JobItem::parameterContainerItem()
{
    return &m_parameterContainer;
}

FitParameterContainerItem* JobItem::fitParameterContainerItem()
{
    if (FitSuiteItem* item = fitSuiteItem())
        return item->fitParameterContainerItem();

    return nullptr;
}

DataItem* JobItem::createNewDataItem()
{
    DataItem* dataItem;
    if (isSpecularJob())
        dataItem = new SpecularDataItem();
    else if (isIntensityJob())
        dataItem = new IntensityDataItem();
    else
        ASSERT(0);

    dataItem->setParentAndModel(this, model());
    return dataItem;
}

void JobItem::createSumulatedDataItem()
{
    ASSERT(!simulatedDataItem());
    m_simulatedDataItem.reset(createNewDataItem());
}

IntensityDataItem* JobItem::intensityDataItem()
{
    return dynamic_cast<IntensityDataItem*>(m_simulatedDataItem.get());
}

DataItem* JobItem::simulatedDataItem()
{
    return m_simulatedDataItem.get();
}

void JobItem::createDiffDataItem()
{
    ASSERT(!diffDataItem());
    m_diffDataItem.reset(createNewDataItem());

    // Immediately create coordinate system and axes units
    GUI::Model::JobItemUtils::setDataItemCoords(diffDataItem(), instrumentItem());
    diffDataItem()->updateCoords(instrumentItem());

    if (isSpecularJob())
        dynamic_cast<SpecularDataItem*>(diffDataItem())->setDiffPlotStyle();
}

DataItem* JobItem::diffDataItem()
{
    return m_diffDataItem.get();
}

RealItem* JobItem::createRealItem()
{
    ASSERT(!realItem());
    m_realItem = std::make_unique<RealItem>();
    m_realItem->setParentAndModel(this, model());

    return m_realItem.get();
}

void JobItem::copyRealItemIntoJob(const RealItem* srcRealItem)
{
    createRealItem();
    srcRealItem->copyTo(realItem());

    if (isSpecularJob())
        m_realItem->specularDataItem()->setRealPlotStyle();
}

RealItem* JobItem::realItem()
{
    return m_realItem.get();
}

void JobItem::adjustReaDataToJobInstrument()
{
    // Link RealItem to the JobItem's instrument.
    // (re-)Linking is necessary because of following reason
    // 1) Copying of RealItem from RealModel on board of JobItem requires relink to the copied
    //    instrument
    // 2) During relink all masks (if exists) will be converted to the default units of current
    // detector
    realItem()->linkToInstrument(instrumentItem());

    // Copy masks and ROI from RealItem on board of instrument.
    instrumentItem()->importMasks(realItem()->maskContainerItem());

    // Crop RealItem to the region of interest.
    if (instrumentItem()->is<Instrument2DItem>())
        cropRealData();
}

//! Crops RealItem to the region of interest.
void JobItem::cropRealData()
{
    ASSERT(instrumentItem()->is<Instrument2DItem>());

    // Adjust real data to the size of region of interest
    IntensityDataItem* intensityItem = realItem()->intensityDataItem();

    std::unique_ptr<Datafield> origData(intensityItem->datafield()->clone());

    auto* instrument_item = instrument2DItem();
    GUI::Model::JobItemUtils::createDefaultDetectorMap(intensityItem, instrument_item);

    instrument_item->createInstrument()->detector().iterateOverNonMaskedPoints(
        [&](IDetector::const_iterator it) {
            auto* cropped_data = intensityItem->datafield();
            (*cropped_data)[it.roiIndex()] = (*origData)[it.detectorIndex()];
        });

    intensityItem->updateDataRange();
}

void JobItem::writeNonSessionItems(QXmlStreamWriter* writer) const
{
    GUI::Session::XML::writeAttribute(writer, GUI::Session::XML::Version, int(1));

    // NOTE: The ordering of the XML elements are important in initialization

    // simulation options
    writer->writeStartElement(Tags::SimulationOptions);
    m_simulationOptionsItem.writeContentTo(writer);
    writer->writeEndElement();

    // instrument
    Streamer(writer).write<InstrumentItemCatalog>(Tags::Instrument, m_instrument.get());

    // sample
    writer->writeStartElement(Tags::Sample);
    m_sampleItem.writeContentTo(writer);
    writer->writeEndElement();

    // parameters
    writer->writeStartElement(Tags::ParameterContainer);
    m_parameterContainer.writeContentTo(writer);
    writer->writeEndElement();

    // data members
    writer->writeStartElement(Tags::Identifier);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_identifier);
    writer->writeEndElement();

    writer->writeStartElement(Tags::Activity);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_activity);
    writer->writeEndElement();

    writer->writeStartElement(Tags::PresentationType);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_presentationType);
    writer->writeEndElement();

    writer->writeStartElement(Tags::BeginTime);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_beginTime.toString(Qt::ISODateWithMs));
    writer->writeEndElement();

    writer->writeStartElement(Tags::EndTime);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_endTime.toString(Qt::ISODateWithMs));
    writer->writeEndElement();

    writer->writeStartElement(Tags::Status);
    GUI::Session::XML::writeAttribute(writer, Tags::value, jobStatusToString(m_status));
    writer->writeEndElement();

    writer->writeStartElement(Tags::Progress);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_progress);
    writer->writeEndElement();

    writer->writeStartElement(Tags::Comments);
    GUI::Session::XML::writeAttribute(writer, Tags::value, m_comments);
    writer->writeEndElement();

    // Output Data
    if (m_simulatedDataItem) {
        writer->writeStartElement(Tags::SimulatedData);
        GUI::Session::XML::writeItemAndChildItems(writer, m_simulatedDataItem.get());
        writer->writeEndElement();
    }

    // Real Data
    if (m_realItem) {
        writer->writeStartElement(Tags::RealItem);
        GUI::Session::XML::writeItemAndChildItems(writer, m_realItem.get());
        writer->writeEndElement();
    }

    // Diff Data
    if (m_diffDataItem) {
        writer->writeStartElement(Tags::DiffData);
        GUI::Session::XML::writeItemAndChildItems(writer, m_diffDataItem.get());
        writer->writeEndElement();
    }

    // Fit Suite
    if (m_fitSuiteItem) {
        writer->writeStartElement(Tags::FitSuite);
        GUI::Session::XML::writeItemAndChildItems(writer, m_fitSuiteItem.get());
        writer->writeEndElement();
    }
}


template <class T>
inline T* readItemToEnd(QXmlStreamReader* reader, SessionItem* parent_item, const QString& tag)
{
    T* item = dynamic_cast<T*>(GUI::Session::XML::readItemAndChildItems(reader, parent_item, tag));
    GUI::Session::XML::gotoEndElementOfTag(reader, tag);
    return item;
}

void JobItem::readNonSessionItems(QXmlStreamReader* reader)
{
    const int version = reader->attributes().value(GUI::Session::XML::Version).toInt();

    if (version < 1)
        throw DeserializationException::tooOld();

    if (version > 1)
        throw DeserializationException::tooNew();

    while (reader->readNextStartElement()) {
        if (reader->name() == Tags::SimulationOptions) {
            m_simulationOptionsItem.readContentFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::SimulationOptions);
        } else if (reader->name() == Tags::Instrument) {
            InstrumentItem* p = nullptr;
            Streamer(reader).read<InstrumentItemCatalog>(Tags::Instrument, p);
            m_instrument.reset(p);
        } else if (reader->name() == Tags::ParameterContainer) {
            m_parameterContainer.readContentFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::ParameterContainer);
        } else if (reader->name() == Tags::Sample) {
            m_sampleItem.readContentFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::Sample);
        } else if (reader->name() == Tags::Activity) {
            GUI::Session::XML::readAttribute(reader, Tags::value, &m_activity);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::Activity);
        } else if (reader->name() == Tags::PresentationType) {
            GUI::Session::XML::readAttribute(reader, Tags::value, &m_presentationType);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::PresentationType);
        } else if (reader->name() == Tags::Comments) {
            GUI::Session::XML::readAttribute(reader, Tags::value, &m_comments);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::Comments);
        } else if (reader->name() == Tags::Status) {
            QString status_str;
            GUI::Session::XML::readAttribute(reader, Tags::value, &status_str);
            m_status = jobStatusFromString(status_str);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::Status);
        } else if (reader->name() == Tags::Progress) {
            QString progress_str;
            GUI::Session::XML::readAttribute(reader, Tags::value, &progress_str);
            m_progress = progress_str.toInt();
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::Progress);
        } else if (reader->name() == Tags::BeginTime) {
            QString begin_t_str;
            GUI::Session::XML::readAttribute(reader, Tags::value, &begin_t_str);
            m_beginTime = QDateTime::fromString(begin_t_str, Qt::ISODateWithMs);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::BeginTime);
        } else if (reader->name() == Tags::EndTime) {
            QString end_t_str;
            GUI::Session::XML::readAttribute(reader, Tags::value, &end_t_str);
            m_endTime = QDateTime::fromString(end_t_str, Qt::ISODateWithMs);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::EndTime);
        } else if (reader->name() == Tags::Identifier) {
            GUI::Session::XML::readAttribute(reader, Tags::value, &m_identifier);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tags::Identifier);
        } else if (reader->name() == Tags::SimulatedData) {
            m_simulatedDataItem.reset(readItemToEnd<DataItem>(reader, this, Tags::SimulatedData));
        } else if (reader->name() == Tags::RealItem) {
            m_realItem.reset(readItemToEnd<RealItem>(reader, this, Tags::RealItem));
        } else if (reader->name() == Tags::DiffData) {
            m_diffDataItem.reset(readItemToEnd<DataItem>(reader, this, Tags::DiffData));
        } else if (reader->name() == Tags::FitSuite) {
            m_fitSuiteItem.reset(readItemToEnd<FitSuiteItem>(reader, this, Tags::FitSuite));
        } else {
            reader->skipCurrentElement();
        }
    }
}

//! Updates the name of file to store intensity data.

void JobItem::updateDataFileName()
{
    if (DataItem* item = simulatedDataItem())
        item->setFileName(GUI::Model::FilenameUtils::jobResultsFileName(itemName()));

    if (RealItem* real = realItem()) {
        if (DataItem* item = real->dataItem())
            item->setFileName(GUI::Model::FilenameUtils::jobReferenceFileName(itemName()));

        if (DataItem* item = real->nativeData())
            item->setFileName(GUI::Model::FilenameUtils::jobNativeDataFileName(getIdentifier()));
    }
}
