//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/MinimizerItem.cpp
//! @brief     Implements MinimizerItem class
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/MinimizerItem.h"
#include "Fit/Adapter/GSLLevenbergMarquardtMinimizer.h"
#include "Fit/Adapter/GSLMultiMinimizer.h"
#include "Fit/Adapter/GeneticMinimizer.h"
#include "Fit/Adapter/Minuit2Minimizer.h"
#include "Fit/Adapter/SimAnMinimizer.h"
#include "GUI/Model/CatJob/MinimizerItemCatalog.h"
#include "Sim/Fitting/ObjectiveMetric.h"
#include "Sim/Fitting/ObjectiveMetricUtils.h"

namespace {

const QMap<QString, QString> minimizer_names_map = {
    {"Minuit2", MinuitMinimizerItem::M_TYPE},
    {"GSL MultiMin", GSLMultiMinimizerItem::M_TYPE},
    {"TMVA Genetic", GeneticMinimizerItem::M_TYPE},
    {"GSL Simulated Annealing", SimAnMinimizerItem::M_TYPE},
    {"GSL Levenberg-Marquardt", GSLLMAMinimizerItem::M_TYPE}};
const QString default_minimizer = minimizer_names_map.key(MinuitMinimizerItem::M_TYPE);

#ifdef ALGORITHM_LIST
QMap<QString /*algorithm*/, QString /*minimizer*/> algorithm_minimizer_map;
void add_algorithm_from_minimizer_to_list_and_map(const QString& minimizer,
                                                  QStringList& common_algorithms_list,
                                                  QStringList& common_algorithms_descriptions)
{
    QStringList algorithms = MinimizerItemCatalog::algorithmNames(minimizer);
    QStringList descriptions = MinimizerItemCatalog::algorithmDescriptions(minimizer);
    ASSERT(algorithms.size() == descriptions.size())

    for (int i=0; i<algorithms.size(); i++) {
        QString name = "   " + algorithms[i];
        common_algorithms_list.append(name);
        common_algorithms_descriptions.append(descriptions[i]);
        algorithm_minimizer_map.insert(name, minimizer_names_map.key(minimizer));
    }
}
void create_algorithm_list_and_map(QString& default_common_algorithm,
                                   QStringList& algorithms_list,
                                   QStringList& algorithms_descriptions)
{
    // Group headers may not directly correspond to minimizers, so for them we don't use descriptions from MinimizerInfo

    algorithms_list.clear();
    algorithms_descriptions.clear();

    // group 1
    // group header. Can't be selected and have no mapped minimizer
    algorithms_list.append("ROOT Minuit2");
    algorithms_descriptions.append("Minuit2 minimizer from CERN ROOT package");

    add_algorithm_from_minimizer_to_list_and_map(MinuitMinimizerItem::M_TYPE, algorithms_list, algorithms_descriptions);

    // group 2
    algorithms_list.append("ROOT TMVA");
    algorithms_descriptions.append("Genetic minimizer from TMVA package, part of CERN ROOT ");

    add_algorithm_from_minimizer_to_list_and_map(GeneticMinimizerItem::M_TYPE,
                                                 algorithms_list,
                                                 algorithms_descriptions);

    // group 3
    algorithms_list.append("GSL MultiMin");
    algorithms_descriptions.append("Multidimensional Minimization from GSL library");

    add_algorithm_from_minimizer_to_list_and_map(GSLMultiMinimizerItem::M_TYPE,
                                                 algorithms_list,
                                                 algorithms_descriptions);

    // group 4
    algorithms_list.append("GSL");
    algorithms_descriptions.append("Other minimizers from GSL library");

    add_algorithm_from_minimizer_to_list_and_map(SimAnMinimizerItem::M_TYPE,
                                                 algorithms_list,
                                                 algorithms_descriptions);

    add_algorithm_from_minimizer_to_list_and_map(GSLLMAMinimizerItem::M_TYPE,
                                                 algorithms_list,
                                                 algorithms_descriptions);

    default_common_algorithm = algorithms_list[1]; // after the first header
}
#endif

} // namespace


MinimizerItem::MinimizerItem(const QString& model_type)
    : SessionItem(model_type)
{
}

// ----------------------------------------------------------------------------

MinimizerContainerItem::MinimizerContainerItem()
    : MinimizerItem(M_TYPE)
{
    addProperty<MinuitMinimizerItem>(MinuitMinimizerItem::M_TYPE);
    addProperty<GSLMultiMinimizerItem>(GSLMultiMinimizerItem::M_TYPE);
    addProperty<GeneticMinimizerItem>(GeneticMinimizerItem::M_TYPE);
    addProperty<SimAnMinimizerItem>(SimAnMinimizerItem::M_TYPE);
    addProperty<GSLLMAMinimizerItem>(GSLLMAMinimizerItem::M_TYPE);

#ifdef ALGORITHM_LIST
    QString default_common_algorithm;
    QStringList common_algorithms_list;
    QStringList common_algorithms_descriptions;
    create_algorithm_list_and_map(default_common_algorithm,
                                  common_algorithms_list,
                                  common_algorithms_descriptions);
    ComboProperty common_algorithm_combo =
        ComboProperty::fromList(common_algorithms_list, default_common_algorithm);
    common_algorithm_combo.setToolTips(common_algorithms_descriptions);
    addProperty(P_COMMON_ALGORITHMS, common_algorithm_combo.variant());
#endif

    ComboProperty minimizer_combo =
        ComboProperty::fromList(minimizer_names_map.keys(), default_minimizer);
    addProperty(P_MINIMIZERS, minimizer_combo.variant());

    ComboProperty metric_combo = ComboProperty::fromStdVec(
        ObjectiveMetricUtils::metricNames(), ObjectiveMetricUtils::defaultMetricName());
    addProperty(P_METRIC, metric_combo.variant());

    ComboProperty norm_combo = ComboProperty::fromStdVec(ObjectiveMetricUtils::normNames(),
                                                         ObjectiveMetricUtils::defaultNormName());
    addProperty(P_NORM, norm_combo.variant());
}

MinuitMinimizerItem* MinimizerContainerItem::minimizerItemMinuit() const
{
    return item<MinuitMinimizerItem>(MinuitMinimizerItem::M_TYPE);
}

GSLMultiMinimizerItem* MinimizerContainerItem::minimizerItemGSLMulti() const
{
    return item<GSLMultiMinimizerItem>(GSLMultiMinimizerItem::M_TYPE);
}

GeneticMinimizerItem* MinimizerContainerItem::minimizerItemGenetic() const
{
    return item<GeneticMinimizerItem>(GeneticMinimizerItem::M_TYPE);
}

SimAnMinimizerItem* MinimizerContainerItem::minimizerItemSimAn() const
{
    return item<SimAnMinimizerItem>(SimAnMinimizerItem::M_TYPE);
}

GSLLMAMinimizerItem* MinimizerContainerItem::minimizerItemGSLLMA() const
{
    return item<GSLLMAMinimizerItem>(GSLLMAMinimizerItem::M_TYPE);
}

MinimizerItem* MinimizerContainerItem::currentMinimizerItem() const
{
    QString M_TYPE_name = minimizer_names_map.value(currentMinimizer());
    return item<MinimizerItem>(M_TYPE_name);
}

QString MinimizerContainerItem::currentMinimizerM_TYPE() const
{
    return minimizer_names_map.value(minimizerCombo().currentValue());
}

QString MinimizerContainerItem::currentMinimizer() const
{
    return minimizerCombo().currentValue();
}

void MinimizerContainerItem::setCurrentMinimizer(const QString& name)
{
    ComboProperty combo = minimizerCombo();
    combo.setCurrentValue(name);
    setItemValue(P_MINIMIZERS, combo.variant());
}

ComboProperty MinimizerContainerItem::minimizerCombo() const
{
    return getItemValue(P_MINIMIZERS).value<ComboProperty>();
}
#ifdef ALGORITHM_LIST
bool MinimizerContainerItem::algorithmHasMinimizer(const QString &name)
{
    return algorithm_minimizer_map.contains(name);
}

void MinimizerContainerItem::setCurrentCommonAlgorithm(const QString& name)
{
    ComboProperty combo = commonAlgorithmCombo();
    combo.setCurrentValue(name);
    setItemValue(P_COMMON_ALGORITHMS, combo.variant());

    ASSERT(algorithmHasMinimizer(name));
    setCurrentMinimizer(algorithm_minimizer_map.value(name));
}

ComboProperty MinimizerContainerItem::commonAlgorithmCombo() const
{
    return getItemValue(P_COMMON_ALGORITHMS).value<ComboProperty>();
}
#endif
QString MinimizerContainerItem::currentObjectiveMetric() const
{
    return objectiveMetricCombo().currentValue();
}

void MinimizerContainerItem::setCurrentObjectiveMetric(const QString& name)
{
    ComboProperty combo = objectiveMetricCombo();
    combo.setCurrentValue(name);
    setItemValue(P_METRIC, combo.variant());
}

ComboProperty MinimizerContainerItem::objectiveMetricCombo() const
{
    return getItemValue(P_METRIC).value<ComboProperty>();
}

QString MinimizerContainerItem::currentNormFunction() const
{
    return normFunctionCombo().currentValue();
}

void MinimizerContainerItem::setCurrentNormFunction(const QString& name)
{
    ComboProperty combo = normFunctionCombo();
    combo.setCurrentValue(name);
    setItemValue(P_NORM, combo.variant());
}

ComboProperty MinimizerContainerItem::normFunctionCombo() const
{
    return getItemValue(P_NORM).value<ComboProperty>();
}

std::unique_ptr<IMinimizer> MinimizerContainerItem::createMinimizer() const
{
    return currentMinimizerItem()->createMinimizer();
}

std::unique_ptr<ObjectiveMetric> MinimizerContainerItem::createMetric() const
{
    return ObjectiveMetricUtils::createMetric(currentObjectiveMetric().toStdString(),
                                              currentNormFunction().toStdString());
}

// ----------------------------------------------------------------------------

MinuitMinimizerItem::MinuitMinimizerItem()
    : MinimizerItem(M_TYPE)
{
    ComboProperty algo = MinimizerItemCatalog::algorithmCombo(modelType());
    addProperty(P_ALGORITHMS, algo.variant());

    addProperty(P_STRATEGY, int(1));

    addProperty(P_ERRORDEF, double(1.0));

    addProperty(P_TOLERANCE, double(0.01));

    addProperty(P_PRECISION, double(-1.0));

    addProperty(P_MAXFUNCTIONCALLS, int(0));
}

QString MinuitMinimizerItem::currentAlgorithm() const
{
    return algorithmCombo().currentValue();
}

void MinuitMinimizerItem::setCurrentAlgorithm(const QString& name)
{
    ComboProperty combo = algorithmCombo();
    combo.setCurrentValue(name);
    setItemValue(P_ALGORITHMS, combo.variant());
}

ComboProperty MinuitMinimizerItem::algorithmCombo() const
{
    return getItemValue(P_ALGORITHMS).value<ComboProperty>();
}

int MinuitMinimizerItem::strategy() const
{
    return getItemValue(P_STRATEGY).toInt();
}

void MinuitMinimizerItem::setStrategy(int value)
{
    setItemValue(P_STRATEGY, value);
}

double MinuitMinimizerItem::errorDefinition() const
{
    return getItemValue(P_ERRORDEF).toDouble();
}

void MinuitMinimizerItem::setErrorDefinition(double value)
{
    setItemValue(P_ERRORDEF, value);
}

double MinuitMinimizerItem::tolerance() const
{
    return getItemValue(P_TOLERANCE).toDouble();
}

void MinuitMinimizerItem::setTolerance(double value)
{
    setItemValue(P_TOLERANCE, value);
}

double MinuitMinimizerItem::precision() const
{
    return getItemValue(P_PRECISION).toDouble();
}

void MinuitMinimizerItem::setPrecision(double value)
{
    setItemValue(P_PRECISION, value);
}

int MinuitMinimizerItem::maxFuncCalls() const
{
    return getItemValue(P_MAXFUNCTIONCALLS).toInt();
}

void MinuitMinimizerItem::setMaxFuncCalls(int value)
{
    setItemValue(P_MAXFUNCTIONCALLS, value);
}

std::unique_ptr<IMinimizer> MinuitMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new Minuit2Minimizer(currentAlgorithm().toStdString());
    domainMinimizer->setStrategy(strategy());
    domainMinimizer->setErrorDefinition(errorDefinition());
    domainMinimizer->setTolerance(tolerance());
    domainMinimizer->setPrecision(precision());
    domainMinimizer->setMaxFunctionCalls(maxFuncCalls());

    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

// ----------------------------------------------------------------------------

GSLMultiMinimizerItem::GSLMultiMinimizerItem()
    : MinimizerItem(M_TYPE)
{
    ComboProperty algo = MinimizerItemCatalog::algorithmCombo(modelType());
    addProperty(P_ALGORITHMS, algo.variant());

    addProperty(P_MAXITERATIONS, int(0));
}

QString GSLMultiMinimizerItem::currentAlgorithm() const
{
    return algorithmCombo().currentValue();
}

void GSLMultiMinimizerItem::setCurrentAlgorithm(const QString& name)
{
    ComboProperty combo = algorithmCombo();
    combo.setCurrentValue(name);
    setItemValue(P_ALGORITHMS, combo.variant());
}

ComboProperty GSLMultiMinimizerItem::algorithmCombo() const
{
    return getItemValue(P_ALGORITHMS).value<ComboProperty>();
}

int GSLMultiMinimizerItem::maxIterations() const
{
    return getItemValue(P_MAXITERATIONS).toInt();
}

void GSLMultiMinimizerItem::setMaxIterations(int value)
{
    setItemValue(P_MAXITERATIONS, value);
}

std::unique_ptr<IMinimizer> GSLMultiMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new GSLMultiMinimizer(currentAlgorithm().toStdString());
    domainMinimizer->setMaxIterations(maxIterations());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

// ----------------------------------------------------------------------------

GeneticMinimizerItem::GeneticMinimizerItem()
    : MinimizerItem(M_TYPE)
{
    addProperty(P_TOLERANCE, double(0.01));
    addProperty(P_MAXITERATIONS, int(3));
    addProperty(P_POPULATIONSIZE, int(300));
    addProperty(P_RANDOMSEED, int(0));
}

double GeneticMinimizerItem::tolerance() const
{
    return getItemValue(P_TOLERANCE).toDouble();
}

void GeneticMinimizerItem::setTolerance(double value)
{
    setItemValue(P_TOLERANCE, value);
}

int GeneticMinimizerItem::maxIterations() const
{
    return getItemValue(P_MAXITERATIONS).toInt();
}

void GeneticMinimizerItem::setMaxIterations(int value)
{
    setItemValue(P_MAXITERATIONS, value);
}

int GeneticMinimizerItem::populationSize() const
{
    return getItemValue(P_POPULATIONSIZE).toInt();
}

void GeneticMinimizerItem::setPopulationSize(int value)
{
    setItemValue(P_POPULATIONSIZE, value);
}

int GeneticMinimizerItem::randomSeed() const
{
    return getItemValue(P_RANDOMSEED).toInt();
}

void GeneticMinimizerItem::setRandomSeed(int value)
{
    setItemValue(P_RANDOMSEED, value);
}

std::unique_ptr<IMinimizer> GeneticMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new GeneticMinimizer();
    domainMinimizer->setTolerance(tolerance());
    domainMinimizer->setMaxIterations(maxIterations());
    domainMinimizer->setPopulationSize(populationSize());
    domainMinimizer->setRandomSeed(randomSeed());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

// ----------------------------------------------------------------------------

SimAnMinimizerItem::SimAnMinimizerItem()
    : MinimizerItem(M_TYPE)
{
    addProperty(P_MAXITERATIONS, int(100));
    addProperty(P_ITERATIONSTEMP, int(10));
    addProperty(P_STEPSIZE, double(1.0));
    addProperty(P_BOLTZMANN_K, double(1.0));
    addProperty(P_BOLTZMANN_TINIT, double(50.0));
    addProperty(P_BOLTZMANN_MU, double(1.05));
    addProperty(P_BOLTZMANN_TMIN, double(0.1));
}

int SimAnMinimizerItem::maxIterations() const
{
    return getItemValue(P_MAXITERATIONS).toInt();
}

void SimAnMinimizerItem::setMaxIterations(int value)
{
    setItemValue(P_MAXITERATIONS, value);
}

int SimAnMinimizerItem::iterationsAtEachTemp() const
{
    return getItemValue(P_ITERATIONSTEMP).toInt();
}

void SimAnMinimizerItem::setIterationsAtEachTemp(int value)
{
    setItemValue(P_ITERATIONSTEMP, value);
}

double SimAnMinimizerItem::stepSize() const
{
    return getItemValue(P_STEPSIZE).toDouble();
}

void SimAnMinimizerItem::setStepSize(double value)
{
    setItemValue(P_STEPSIZE, value);
}

double SimAnMinimizerItem::boltzmanK() const
{
    return getItemValue(P_BOLTZMANN_K).toDouble();
}

void SimAnMinimizerItem::setBoltzmanK(double value)
{
    setItemValue(P_BOLTZMANN_K, value);
}

double SimAnMinimizerItem::boltzmanInitT() const
{
    return getItemValue(P_BOLTZMANN_TINIT).toDouble();
}

void SimAnMinimizerItem::setBoltzmanInitT(double value)
{
    setItemValue(P_BOLTZMANN_TINIT, value);
}

double SimAnMinimizerItem::boltzmanMu() const
{
    return getItemValue(P_BOLTZMANN_MU).toDouble();
}

void SimAnMinimizerItem::setBoltzmanMu(double value)
{
    setItemValue(P_BOLTZMANN_MU, value);
}

double SimAnMinimizerItem::boltzmanMinT() const
{
    return getItemValue(P_BOLTZMANN_TMIN).toDouble();
}

void SimAnMinimizerItem::setBoltzmanMinT(double value)
{
    setItemValue(P_BOLTZMANN_TMIN, value);
}

std::unique_ptr<IMinimizer> SimAnMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new SimAnMinimizer();
    domainMinimizer->setMaxIterations(maxIterations());
    domainMinimizer->setIterationsAtEachTemp(iterationsAtEachTemp());
    domainMinimizer->setStepSize(stepSize());
    domainMinimizer->setBoltzmannK(boltzmanK());
    domainMinimizer->setBoltzmannInitialTemp(boltzmanInitT());
    domainMinimizer->setBoltzmannMu(boltzmanMu());
    domainMinimizer->setBoltzmannMinTemp(boltzmanMinT());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

// ----------------------------------------------------------------------------

GSLLMAMinimizerItem::GSLLMAMinimizerItem()
    : MinimizerItem(M_TYPE)
{
    addProperty(P_TOLERANCE, double(0.01));
    addProperty(P_MAXITERATIONS, int(0));
}

double GSLLMAMinimizerItem::tolerance() const
{
    return getItemValue(P_TOLERANCE).toDouble();
}

void GSLLMAMinimizerItem::setTolerance(double value)
{
    setItemValue(P_TOLERANCE, value);
}

int GSLLMAMinimizerItem::maxIterations() const
{
    return getItemValue(P_MAXITERATIONS).toInt();
}

void GSLLMAMinimizerItem::setMaxIterations(int value)
{
    setItemValue(P_MAXITERATIONS, value);
}

std::unique_ptr<IMinimizer> GSLLMAMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new GSLLevenbergMarquardtMinimizer();
    domainMinimizer->setTolerance(tolerance());
    domainMinimizer->setMaxIterations(maxIterations());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}
