//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/ApplicationModels.cpp
//! @brief     Defines class holding all application models
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/ApplicationModels.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Model/RealModel.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Util/DeserializationException.h"
#include <QXmlStreamWriter>

ApplicationModels::ApplicationModels()
    : m_realDataModel(nullptr)
    , m_jobModel(nullptr)
{
    //! creates and initializes models, order is important
    m_realDataModel = new RealModel(this);
    m_jobModel = new JobModel(this);

    connectModel(m_realDataModel);
    connectModel(m_jobModel);
}

RealModel* ApplicationModels::realModel() const
{
    return m_realDataModel;
}

JobModel* ApplicationModels::jobModel() const
{
    return m_jobModel;
}

//! Writes all model in file one by one

void ApplicationModels::writeTo(QXmlStreamWriter* writer)
{
    m_realDataModel->writeTo(writer);
    m_jobModel->writeTo(writer);
}

void ApplicationModels::readFrom(QXmlStreamReader* reader, MessageService* messageService)
{
    try {
        if (m_realDataModel->getModelTag() == reader->name())
            m_realDataModel->readFrom(reader, messageService);
        else if (m_jobModel->getModelTag() == reader->name())
            m_jobModel->readFrom(reader, messageService);
    } catch (DeserializationException& ex) {
        m_realDataModel->clear();
        m_jobModel->clear();
        reader->raiseError(ex.text());
    }
}


QVector<SessionItem*> ApplicationModels::nonXMLItems() const
{
    ASSERT(m_realDataModel && m_jobModel);
    return QVector<SessionItem*>() << m_realDataModel->nonXMLItems() << m_jobModel->nonXMLItems();
}

void ApplicationModels::connectModel(SessionModel* model) const
{
    if (!model)
        return;
    connect(model, &SessionModel::dataChanged, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
    connect(model, &SessionModel::rowsRemoved, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
    connect(model, &SessionModel::rowsInserted, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
}
