//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/FormFactorItems.h
//! @brief     Defines FormFactorItems classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_SAMPLE_FORMFACTORITEMS_H
#define BORNAGAIN_GUI_MODEL_SAMPLE_FORMFACTORITEMS_H

#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Support/XML/Serialize.h"
#include <QVector>
#include <memory>

class IFormFactor;
class GroupInfo;

#define FF_PROPERTY(nameLower, nameUpper)                                                          \
private:                                                                                           \
    DoubleProperty m_##nameLower;                                                                  \
                                                                                                   \
public:                                                                                            \
    DoubleDescriptor nameLower() const { return m_##nameLower; }                                   \
    void set##nameUpper(double v) { m_##nameLower.set(v); }

class FormFactorItem {
public:
    virtual ~FormFactorItem() = default;

public:
    virtual std::unique_ptr<IFormFactor> createFormFactor() const = 0;
    virtual DoubleDescriptors geometryValues() const = 0;
    virtual QVector<DoubleProperty*> serializationProperties() = 0;
    virtual void serialize(Streamer& s);
};

class Pyramid2Item : public FormFactorItem {
public:
    Pyramid2Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoubleDescriptors geometryValues() const override
    {
        return {length(), width(), height(), alpha()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height, &m_alpha};
    }
};

class BarGaussItem : public FormFactorItem {
public:
    BarGaussItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {length(), width(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height};
    }
};

class BarLorentzItem : public FormFactorItem {
public:
    BarLorentzItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {length(), width(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height};
    }
};

class BoxItem : public FormFactorItem {
public:
    BoxItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {length(), width(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height};
    }
};

class ConeItem : public FormFactorItem {
public:
    ConeItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoubleDescriptors geometryValues() const override { return {radius(), height(), alpha()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_radius, &m_height, &m_alpha};
    }
};

class Pyramid6Item : public FormFactorItem {
public:
    Pyramid6Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoubleDescriptors geometryValues() const override { return {baseEdge(), height(), alpha()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_baseEdge, &m_height, &m_alpha};
    }
};

class Bipyramid4Item : public FormFactorItem {
public:
    Bipyramid4Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(heightRatio, HeightRatio)
    FF_PROPERTY(alpha, Alpha)

    DoubleDescriptors geometryValues() const override
    {
        return {length(), height(), heightRatio(), alpha()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_height, &m_heightRatio, &m_alpha};
    }
};

class CylinderItem : public FormFactorItem {
public:
    CylinderItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {radius(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_radius, &m_height}; }
};

class DodecahedronItem : public FormFactorItem {
public:
    DodecahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(edge, Edge)

    DoubleDescriptors geometryValues() const override { return {edge()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_edge}; }
};

class EllipsoidalCylinderItem : public FormFactorItem {
public:
    EllipsoidalCylinderItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radiusX, RadiusX)
    FF_PROPERTY(radiusY, RadiusY)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {radiusX(), radiusY(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_radiusX, &m_radiusY, &m_height};
    }
};

class SphereItem : public FormFactorItem {
public:
    SphereItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)

    DoubleDescriptors geometryValues() const override { return {radius()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_radius}; }
};

class SpheroidItem : public FormFactorItem {
public:
    SpheroidItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {radius(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_radius, &m_height}; }
};

class HemiEllipsoidItem : public FormFactorItem {
public:
    HemiEllipsoidItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radiusX, RadiusX)
    FF_PROPERTY(radiusY, RadiusY)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {radiusX(), radiusY(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_radiusX, &m_radiusY, &m_height};
    }
};

class IcosahedronItem : public FormFactorItem {
public:
    IcosahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(edge, Edge)

    DoubleDescriptors geometryValues() const override { return {edge()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_edge}; }
};

class Prism3Item : public FormFactorItem {
public:
    Prism3Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {baseEdge(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_baseEdge, &m_height}; }
};

class Prism6Item : public FormFactorItem {
public:
    Prism6Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {baseEdge(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_baseEdge, &m_height}; }
};

class Pyramid4Item : public FormFactorItem {
public:
    Pyramid4Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoubleDescriptors geometryValues() const override { return {baseEdge(), height(), alpha()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_baseEdge, &m_height, &m_alpha};
    }
};

class CosineRippleBoxItem : public FormFactorItem {
public:
    CosineRippleBoxItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {length(), width(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height};
    }
};

class CosineRippleGaussItem : public FormFactorItem {
public:
    CosineRippleGaussItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {length(), width(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height};
    }
};

class CosineRippleLorentzItem : public FormFactorItem {
public:
    CosineRippleLorentzItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)

    DoubleDescriptors geometryValues() const override { return {length(), width(), height()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height};
    }
};

class SawtoothRippleBoxItem : public FormFactorItem {
public:
    SawtoothRippleBoxItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(asymmetry, Asymmetry)

    DoubleDescriptors geometryValues() const override
    {
        return {length(), width(), height(), asymmetry()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height, &m_asymmetry};
    }
};

class SawtoothRippleGaussItem : public FormFactorItem {
public:
    SawtoothRippleGaussItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(asymmetry, Asymmetry)

    DoubleDescriptors geometryValues() const override
    {
        return {length(), width(), height(), asymmetry()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height, &m_asymmetry};
    }
};

class SawtoothRippleLorentzItem : public FormFactorItem {
public:
    SawtoothRippleLorentzItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(width, Width)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(asymmetry, Asymmetry)

    DoubleDescriptors geometryValues() const override
    {
        return {length(), width(), height(), asymmetry()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_width, &m_height, &m_asymmetry};
    }
};

class Pyramid3Item : public FormFactorItem {
public:
    Pyramid3Item();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(baseEdge, BaseEdge)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(alpha, Alpha)

    DoubleDescriptors geometryValues() const override { return {baseEdge(), height(), alpha()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_baseEdge, &m_height, &m_alpha};
    }
};

class TruncatedCubeItem : public FormFactorItem {
public:
    TruncatedCubeItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(removedLength, RemovedLength)

    DoubleDescriptors geometryValues() const override { return {length(), removedLength()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_removedLength};
    }
};

class TruncatedSphereItem : public FormFactorItem {
public:
    TruncatedSphereItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(removedTop, RemovedTop)

    DoubleDescriptors geometryValues() const override { return {radius(), height(), removedTop()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_radius, &m_height, &m_removedTop};
    }
};

class TruncatedSpheroidItem : public FormFactorItem {
public:
    TruncatedSpheroidItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(height, Height)
    FF_PROPERTY(heightFlattening, HeightFlattening)
    FF_PROPERTY(removedTop, RemovedTop)

    DoubleDescriptors geometryValues() const override
    {
        return {radius(), height(), heightFlattening(), removedTop()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_radius, &m_height, &m_heightFlattening, &m_removedTop};
    }
};

class CantellatedCubeItem : public FormFactorItem {
public:
    CantellatedCubeItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(length, Length)
    FF_PROPERTY(removedLength, RemovedLength)

    DoubleDescriptors geometryValues() const override { return {length(), removedLength()}; }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_length, &m_removedLength};
    }
};

class HorizontalCylinderItem : public FormFactorItem {
public:
    HorizontalCylinderItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(radius, Radius)
    FF_PROPERTY(length, Length)
    FF_PROPERTY(sliceBottom, SliceBottom)
    FF_PROPERTY(sliceTop, SliceTop)

    DoubleDescriptors geometryValues() const override
    {
        return {radius(), length(), sliceBottom(), sliceTop()};
    }
    QVector<DoubleProperty*> serializationProperties() override
    {
        return {&m_radius, &m_length, &m_sliceBottom, &m_sliceTop};
    }
};

class PlatonicOctahedronItem : public FormFactorItem {
public:
    PlatonicOctahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(edge, Edge)

    DoubleDescriptors geometryValues() const override { return {edge()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_edge}; }
};

class PlatonicTetrahedronItem : public FormFactorItem {
public:
    PlatonicTetrahedronItem();
    std::unique_ptr<IFormFactor> createFormFactor() const override;

    FF_PROPERTY(edge, Edge)

    DoubleDescriptors geometryValues() const override { return {edge()}; }
    QVector<DoubleProperty*> serializationProperties() override { return {&m_edge}; }
};

#endif // BORNAGAIN_GUI_MODEL_SAMPLE_FORMFACTORITEMS_H
