//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/SampleValidator.cpp
//! @brief     Implements class SampleValidator
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/SampleValidator.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"


void SampleValidator::validateItem(const ItemWithParticles* item)
{
    if (const auto* p = dynamic_cast<const ParticleCoreShellItem*>(item)) {
        if (!p->core())
            addMessage("Sim/shell particle doesn't have core defined.");
        if (!p->shell())
            addMessage("Sim/shell particle doesn't have shell defined.");
    } else if (const auto* p = dynamic_cast<const ParticleCompositionItem*>(item))
        if (p->particles().isEmpty())
            addMessage("Particle composition doesn't have any particles.");
}

void SampleValidator::validateInterferences(const LayerItem* layer)
{
    ASSERT(layer);
#ifdef WIN32

    int iLayout = 0;
    for (const auto* layout : layer->layouts()) {
        iLayout++;
        if (const auto* c = dynamic_cast<const Interference2DAbstractLatticeItem*>(
                layout->interference().currentItem()))
            if (c->xiIntegration()) {
                const QString message =
                    QString("Interference function of layer '%1', Particle layout %2: "
                            "Integration over Xi is currently not supported "
                            "under Windows")
                        .arg(layer->layerName())
                        .arg(iLayout);
                addMessage(message);
            }
    }
#endif
}

void SampleValidator::addMessage(const QString& m)
{
    m_messages += QString("* ") + m + "\n";
}

bool SampleValidator::isValidMultiLayer(const MultiLayerItem* sample)
{
    m_messages.clear();

    QVector<LayerItem*> layers = sample->layers();

    if (layers.isEmpty())
        addMessage("MultiLayer should contain at least one layer.");
    if (layers.size() == 1)
        if (layers.front()->layouts().isEmpty())
            addMessage("The single layer in your MultiLayer should contain a particle layout.");

    for (const auto* layer : layers) {
        validateInterferences(layer);
        for (const auto* layout : layer->layouts()) {
            if (layout->particles().isEmpty())
                addMessage("Particle layout doesn't contain any particles.");
            for (const auto* particle : layout->containedItemsWithParticles()) {
                validateItem(particle);
            }
        }
    }
    if (!m_messages.isEmpty())
        m_messages = "Can't setup DWBA simulation for given MultiLayer.\n" + m_messages;

    return m_messages.isEmpty();
}
