//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Realspace/RealSpaceMesoCrystalUtils.cpp
//! @brief     Defines GUI::RealSpace::BuilderUtils namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Realspace/RealSpaceMesoCrystalUtils.h"
#include "Base/Const/Units.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/View/Realspace/Particle3DContainer.h"
#include "GUI/View/Realspace/RealSpaceBuilderUtils.h"
#include "GUI/View/Realspace/TransformTo3D.h"
#include "Sample/HardParticle/HardParticles.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"

namespace {

const int n = 10; // TODO: Adjust this parameter based on the size of the mesocrystal

bool isPositionInsideMesoCrystal(const IFormFactor* outerShape, R3 positionInside)
{
    bool check(false);
    if (const auto* ff_Pyramid2 = dynamic_cast<const Pyramid2*>(outerShape)) {
        double L = ff_Pyramid2->length();
        double W = ff_Pyramid2->width();
        double H = ff_Pyramid2->height();
        double alpha = ff_Pyramid2->alpha();

        double l_z =
            L / 2
            - positionInside.z() / std::tan(alpha); // half-length of rectangle at a given height
        double w_z =
            W / 2
            - positionInside.z() / std::tan(alpha); // half-width of rectangle at a given height
        if (std::abs(positionInside.x()) <= l_z && std::abs(positionInside.y()) <= w_z
            && (positionInside.z() >= 0 && positionInside.z() <= H))
            check = true;
    } else if (const auto* ff_BarGauss = dynamic_cast<const BarGauss*>(outerShape)) {
        double L = ff_BarGauss->length();
        double W = ff_BarGauss->width();
        double H = ff_BarGauss->height();

        if (std::abs(positionInside.x()) <= L / 2 && std::abs(positionInside.y()) <= W / 2
            && (positionInside.z() >= 0 && positionInside.z() <= H))
            check = true;
    } else if (const auto* ff_BarLorentz = dynamic_cast<const BarLorentz*>(outerShape)) {
        double L = ff_BarLorentz->length();
        double W = ff_BarLorentz->width();
        double H = ff_BarLorentz->height();

        if (std::abs(positionInside.x()) <= L / 2 && std::abs(positionInside.y()) <= W / 2
            && (positionInside.z() >= 0 && positionInside.z() <= H))
            check = true;
    } else if (const auto* ff_Box = dynamic_cast<const Box*>(outerShape)) {
        double L = ff_Box->length();
        double W = ff_Box->width();
        double H = ff_Box->height();

        if (std::abs(positionInside.x()) <= L / 2 && std::abs(positionInside.y()) <= W / 2
            && (positionInside.z() >= 0 && positionInside.z() <= H))
            check = true;
    } else if (const auto* ff_Cone = dynamic_cast<const Cone*>(outerShape)) {
        double R = ff_Cone->radius();
        double H = ff_Cone->height();
        double alpha = ff_Cone->alpha();

        if (std::abs(positionInside.x()) > R || std::abs(positionInside.y()) > R
            || positionInside.z() < 0 || positionInside.z() > H)
            return check;

        double R_z = R - positionInside.z() / std::tan(alpha);
        if (std::pow(positionInside.x() / R_z, 2) + std::pow(positionInside.y() / R_z, 2) <= 1)
            check = true;
    } else if (const auto* ff_Pyramid6 = dynamic_cast<const Pyramid6*>(outerShape)) {
        double B = ff_Pyramid6->baseEdge();
        double H = ff_Pyramid6->height();
        double alpha = ff_Pyramid6->alpha();

        if (std::abs(positionInside.x()) > B || std::abs(positionInside.y()) > B
            || positionInside.z() < 0 || positionInside.z() > H)
            return check;

        double l_z = B - positionInside.z() / std::tan(alpha); // edge of hexagon at a given height
        double theta_prime = 0; // angle between positionInside & x-axis in positionInside.z() plane
        if (positionInside.x() != 0 || positionInside.y() != 0)
            theta_prime = Units::rad2deg(std::asin(
                std::abs(positionInside.y())
                / std::sqrt(std::pow(positionInside.x(), 2) + std::pow(positionInside.y(), 2))));
        int c = static_cast<int>(theta_prime / 60); // multiplication constant
        double theta = Units::deg2rad(theta_prime - c * 60);
        double k_z = l_z / (std::cos(theta) + std::sin(theta) / std::tan(M_PI / 3));

        if (std::pow(positionInside.x(), 2) + std::pow(positionInside.y(), 2) <= std::pow(k_z, 2))
            check = true;
    } else if (const auto* ff_Bipyramid4 = dynamic_cast<const Bipyramid4*>(outerShape)) {
        double L = ff_Bipyramid4->length();
        double H = ff_Bipyramid4->height();
        double rH = ff_Bipyramid4->heightRatio();
        double alpha = ff_Bipyramid4->alpha();

        double total_Height = H + rH * H;

        if (std::abs(positionInside.x()) > L / 2 || std::abs(positionInside.y()) > L / 2
            || positionInside.z() < 0 || positionInside.z() > total_Height)
            return check;

        // half-length of square (i.e. horizontal cross-section of Bipyramid4) at a given height
        double l_z = L / 2 - std::abs(H - positionInside.z()) / std::tan(alpha);
        if (std::abs(positionInside.x()) <= l_z && std::abs(positionInside.y()) <= l_z)
            check = true;
    } else if (const auto* ff_Cylinder = dynamic_cast<const Cylinder*>(outerShape)) {
        double R = ff_Cylinder->radius();
        double H = ff_Cylinder->height();

        if (std::abs(positionInside.x()) > R || std::abs(positionInside.y()) > R
            || positionInside.z() < 0 || positionInside.z() > H)
            return check;

        if (std::pow(positionInside.x() / R, 2) + std::pow(positionInside.y() / R, 2) <= 1)
            check = true;
    } else if (dynamic_cast<const Dodecahedron*>(outerShape)) {
        // TODO: Implement Dodecahedron
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Dodecahedron not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (const auto* ff_EllipsoidalCylinder =
                   dynamic_cast<const EllipsoidalCylinder*>(outerShape)) {
        double a = ff_EllipsoidalCylinder->radiusX(); // semi-axis length along x
        double b = ff_EllipsoidalCylinder->radiusY(); // semi-axis length along y
        double H = ff_EllipsoidalCylinder->height();

        if (std::abs(positionInside.x()) > a || std::abs(positionInside.y()) > b
            || positionInside.z() < 0 || positionInside.z() > H)
            return check;

        if (std::pow(positionInside.x() / a, 2) + std::pow(positionInside.y() / b, 2) <= 1)
            check = true;
    } else if (const auto* ff_Sphere = dynamic_cast<const Sphere*>(outerShape)) {
        double R = ff_Sphere->radius();

        if (std::abs(positionInside.x()) > R || std::abs(positionInside.y()) > R
            || positionInside.z() < 0 || positionInside.z() > 2 * R)
            return check;

        if (std::pow(positionInside.x() / R, 2) + std::pow(positionInside.y() / R, 2)
                + std::pow((positionInside.z() - R) / R, 2)
            <= 1)
            check = true;
    } else if (const auto* ff_Spheroid = dynamic_cast<const Spheroid*>(outerShape)) {
        double a = ff_Spheroid->radius(); // semi-axis length along x and y
        double H = ff_Spheroid->height();
        double c = H / 2; // semi-axis length along z

        if (std::abs(positionInside.x()) > a || std::abs(positionInside.y()) > a
            || positionInside.z() < 0 || positionInside.z() > H)
            return check;

        if (std::pow(positionInside.x() / a, 2) + std::pow(positionInside.y() / a, 2)
                + std::pow((positionInside.z() - c) / c, 2)
            <= 1)
            check = true;
    } else if (const auto* ff_HemiEllipsoid = dynamic_cast<const HemiEllipsoid*>(outerShape)) {
        double a = ff_HemiEllipsoid->radiusX(); // semi-axis length along x
        double b = ff_HemiEllipsoid->radiusY(); // semi-axis length along y
        double c = ff_HemiEllipsoid->height();  // semi-axis length along z

        if (std::abs(positionInside.x()) > a || std::abs(positionInside.y()) > b
            || positionInside.z() < 0 || positionInside.z() > c)
            return check;

        if (std::pow(positionInside.x() / a, 2) + std::pow(positionInside.y() / b, 2)
                + std::pow(positionInside.z() / c, 2)
            <= 1)
            check = true;
    } else if (dynamic_cast<const Icosahedron*>(outerShape)) {
        // TODO: Implement Icosahedron
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Icosahedron not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (const auto* ff_Prism3 = dynamic_cast<const Prism3*>(outerShape)) {
        double B = ff_Prism3->baseEdge();
        double H = ff_Prism3->height();

        double l = B * std::sin(M_PI / 3);
        double x_shift = B / 2 * std::tan(M_PI / 6);

        if (positionInside.x() + x_shift < 0 || positionInside.x() + x_shift > l
            || std::abs(positionInside.y()) > B / 2 || positionInside.z() < 0
            || positionInside.z() > H)
            return check;

        double theta = 0; // angle between positionInside & x-axis in positionInside.z() plane
        if (positionInside.x() + x_shift != 0 || positionInside.y() != 0)
            theta = std::asin(std::abs(positionInside.y())
                              / std::sqrt(std::pow(positionInside.x() + x_shift, 2)
                                          + std::pow(positionInside.y(), 2)));

        double k = l / (std::sin(theta) / std::tan(M_PI / 6) + std::cos(theta));

        if (std::pow(positionInside.x() + x_shift, 2) + std::pow(positionInside.y(), 2)
            <= std::pow(k, 2))
            check = true;
    } else if (const auto* ff_Prism6 = dynamic_cast<const Prism6*>(outerShape)) {
        double B = ff_Prism6->baseEdge();
        double H = ff_Prism6->height();

        if (std::abs(positionInside.x()) > B || std::abs(positionInside.y()) > B
            || positionInside.z() < 0 || positionInside.z() > H)
            return check;

        double theta_prime = 0; // angle between positionInside & x-axis in positionInside.z() plane
        if (positionInside.x() != 0 || positionInside.y() != 0)
            theta_prime = Units::rad2deg(std::asin(
                std::abs(positionInside.y())
                / std::sqrt(std::pow(positionInside.x(), 2) + std::pow(positionInside.y(), 2))));
        int c = static_cast<int>(theta_prime / 60); // multiplicative constant
        double theta = Units::deg2rad(theta_prime - c * 60);
        double k_z = B / (std::cos(theta) + std::sin(theta) / std::tan(M_PI / 3));

        if (std::pow(positionInside.x(), 2) + std::pow(positionInside.y(), 2) <= std::pow(k_z, 2))
            check = true;
    } else if (const auto* ff_Pyramid4 = dynamic_cast<const Pyramid4*>(outerShape)) {
        double B = ff_Pyramid4->baseEdge();
        double H = ff_Pyramid4->height();
        double alpha = ff_Pyramid4->alpha();

        double l_z =
            B / 2 - positionInside.z() / std::tan(alpha); // half-length of square at a given height
        if (std::abs(positionInside.x()) <= l_z && std::abs(positionInside.y()) <= l_z
            && (positionInside.z() >= 0 && positionInside.z() <= H))
            check = true;
    } else if (dynamic_cast<const CosineRippleBox*>(outerShape)) {
        // TODO: Implement CosineRippleBox
        std::ostringstream ostr;
        ostr << "Sorry, outer shape CosineRippleBox not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const CosineRippleGauss*>(outerShape)) {
        // TODO: Implement CosineRippleGauss
        std::ostringstream ostr;
        ostr << "Sorry, outer shape CosineRippleGauss not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const CosineRippleLorentz*>(outerShape)) {
        // TODO: Implement CosineRippleLorentz
        std::ostringstream ostr;
        ostr << "Sorry, outer shape CosineRippleLorentz not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const SawtoothRippleBox*>(outerShape)) {
        // TODO: Implement SawtoothRippleBox
        std::ostringstream ostr;
        ostr << "Sorry, outer shape SawtoothRippleBox not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const SawtoothRippleGauss*>(outerShape)) {
        // TODO: Implement SawtoothRippleGauss
        std::ostringstream ostr;
        ostr << "Sorry, outer shape SawtoothRippleGauss not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const SawtoothRippleLorentz*>(outerShape)) {
        // TODO: Implement SawtoothRippleLorentz
        std::ostringstream ostr;
        ostr << "Sorry, outer shape SawtoothRippleLorentz not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (const auto* ff_Pyramid3 = dynamic_cast<const Pyramid3*>(outerShape)) {
        double B = ff_Pyramid3->baseEdge();
        double H = ff_Pyramid3->height();
        double alpha = ff_Pyramid3->alpha();

        double B_z =
            B - positionInside.z() * 2 / std::tan(alpha); // edge of triangle at a given height

        double l = B_z * std::sin(M_PI / 3);
        double x_shift = B_z / 2 * std::tan(M_PI / 6);

        if (positionInside.x() + x_shift < 0 || positionInside.x() + x_shift > l
            || std::abs(positionInside.y()) > B_z / 2 || positionInside.z() < 0
            || positionInside.z() > H)
            return check;

        double theta = 0; // angle between positionInside & x-axis in positionInside.z() plane
        if (positionInside.x() + x_shift != 0 || positionInside.y() != 0)
            theta = std::asin(std::abs(positionInside.y())
                              / std::sqrt(std::pow(positionInside.x() + x_shift, 2)
                                          + std::pow(positionInside.y(), 2)));

        double k = l / (std::sin(theta) / std::tan(M_PI / 6) + std::cos(theta));

        if (std::pow(positionInside.x() + x_shift, 2) + std::pow(positionInside.y(), 2)
            <= std::pow(k, 2))
            check = true;
    } else if (dynamic_cast<const TruncatedCube*>(outerShape)) {
        // TODO: Implement Truncated cube
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Truncated cube not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (const auto* ff_TruncatedSphere = dynamic_cast<const TruncatedSphere*>(outerShape)) {
        double R = ff_TruncatedSphere->radius();
        double H = ff_TruncatedSphere->height();
        double deltaH = ff_TruncatedSphere->removedTop();
        if (std::abs(positionInside.x()) > R || std::abs(positionInside.y()) > R
            || positionInside.z() < 0 || positionInside.z() > (H - deltaH))
            return check;

        if (std::pow(positionInside.x() / R, 2) + std::pow(positionInside.y() / R, 2)
                + std::pow((positionInside.z() - (H - R)) / R, 2)
            <= 1)
            check = true;
    } else if (dynamic_cast<const TruncatedSpheroid*>(outerShape)) {
        // TODO: Implement Truncated spheroid
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Truncated spheroid not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const CantellatedCube*>(outerShape)) {
        // TODO: Implement Cantellated cube
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Cantellated cube not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const HorizontalCylinder*>(outerShape)) {
        // TODO: Implement Horizontal cylinder
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Horizontal cylinder not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const PlatonicOctahedron*>(outerShape)) {
        // TODO: Implement Platonic octahedron
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Platonic octahedron not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else if (dynamic_cast<const PlatonicTetrahedron*>(outerShape)) {
        // TODO: Implement Platonic tetrahedron
        std::ostringstream ostr;
        ostr << "Sorry, outer shape Platonic tetrahedron not yet implemented for Mesocrystal";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    }
    return check;
}

} // namespace


RealSpaceMesoCrystal::RealSpaceMesoCrystal(
    const MesoCrystalItem* mesoCrystalItem, double total_abundance, const QVector3D& origin,
    std::function<QColor(const QString&)> fnColorFromMaterialName)
{
    m_mesoCrystalItem = mesoCrystalItem;
    m_total_abundance = total_abundance;
    m_origin = origin;
    m_builderUtils = std::make_unique<GUI::RealSpace::BuilderUtils>(fnColorFromMaterialName);
}

Particle3DContainer RealSpaceMesoCrystal::populateMesoCrystal()
{
    auto mesoCrystal = m_mesoCrystalItem->createMesoCrystal();

    std::unique_ptr<MesoCrystal> M_clone(mesoCrystal->clone()); // clone of the mesoCrystal

    // These methods DO NOT add rotation/translation of the mesoCrystal to its children
    // and hence they need to be added manually
    auto lattice = m_mesoCrystalItem->getLattice();
    auto particleBasis = m_mesoCrystalItem->getBasis();
    auto outerShapeff = m_mesoCrystalItem->getOuterShape();

    const auto* mesoCrystal_rotation = M_clone->rotation();
    auto mesoCrystal_translation = M_clone->particlePosition();

    Particle3DContainer mesoCrystalBasis3DContainer;

    if (dynamic_cast<const ParticleComposition*>(particleBasis.get())) {
        const auto* particleComposition =
            dynamic_cast<const ParticleComposition*>(particleBasis.get());
        mesoCrystalBasis3DContainer =
            m_builderUtils->particleComposition3DContainer(*particleComposition, 1.0, m_origin);
    } else if (dynamic_cast<const ParticleCoreShell*>(particleBasis.get())) {
        const auto* particleCoreShell = dynamic_cast<const ParticleCoreShell*>(particleBasis.get());
        mesoCrystalBasis3DContainer =
            m_builderUtils->particleCoreShell3DContainer(*particleCoreShell, 1.0, m_origin);
    } else if (dynamic_cast<const MesoCrystal*>(particleBasis.get())) {
        // TODO: Implement method to populate MesoCrystal from CORE and NOT from MesoCrystalItem
        // as it is done currently in GUI::RealSpace::BuilderUtils::mesoCrystal3DContainer
        std::ostringstream ostr;
        ostr << "Sorry, MesoCrystal inside MesoCrystal not yet implemented";
        ostr << "\n\nStay tuned!";
        throw std::runtime_error(ostr.str());
    } else {
        const auto* particle = dynamic_cast<const Particle*>(particleBasis.get());
        mesoCrystalBasis3DContainer =
            m_builderUtils->singleParticle3DContainer(*particle, 1.0, m_origin);
    }

    Particle3DContainer mesoCrystal3DContainer;

    for (int k = -n; k <= n; k++) {
        for (int j = -n; j <= n; j++) {
            for (int i = -n; i <= n; i++) {
                auto positionInside = i * lattice.basisVectorA() + j * lattice.basisVectorB()
                                      + k * lattice.basisVectorC();

                if (isPositionInsideMesoCrystal(outerShapeff.get(), positionInside)) {
                    for (size_t it = 0; it < mesoCrystalBasis3DContainer.containerSize(); ++it) {
                        auto particle3D = mesoCrystalBasis3DContainer.createParticle(it);
                        particle3D->addTranslation(
                            QVector3D(static_cast<float>(positionInside.x()),
                                      static_cast<float>(positionInside.y()),
                                      static_cast<float>(positionInside.z())));

                        particle3D->addExtrinsicRotation(
                            m_builderUtils->implementParticleRotationfromIRotation(
                                mesoCrystal_rotation));

                        particle3D->addTranslation(
                            QVector3D(static_cast<float>(mesoCrystal_translation.x()),
                                      static_cast<float>(mesoCrystal_translation.y()),
                                      static_cast<float>(mesoCrystal_translation.z())));

                        mesoCrystal3DContainer.addParticle(
                            particle3D.release(), mesoCrystalBasis3DContainer.particle3DBlend(it));
                    }
                }
            }
        }
    }

    // Add outer shape for visualization
    auto outerShape3D = GUI::View::TransformTo3D::createParticlefromFormfactor(outerShapeff.get());
    outerShape3D->addTransform(
        m_builderUtils->implementParticleRotationfromIRotation(mesoCrystal_rotation),
        QVector3D(static_cast<float>(mesoCrystal_translation.x()),
                  static_cast<float>(mesoCrystal_translation.y()),
                  static_cast<float>(mesoCrystal_translation.z())));

    // assign grey (default) color to the outer shape
    QColor color = {};
    color.setAlphaF(0.3);
    outerShape3D->color = color;
    mesoCrystal3DContainer.addParticle(outerShape3D.release(), true);

    // set the correct abundance for the MesoCrystal
    mesoCrystal3DContainer.setCumulativeAbundance(M_clone->abundance() / m_total_abundance);
    mesoCrystal3DContainer.setParticleType("MesoCrystal");

    return mesoCrystal3DContainer;
}
