//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Tool/WidgetUtils.h
//! @brief     Defines GUI::Util namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_TOOL_WIDGETUTILS_H
#define BORNAGAIN_GUI_VIEW_TOOL_WIDGETUTILS_H

#include "GUI/Model/Descriptor/SelectionDescriptor.h"
#include "GUI/Support/Type/Unit.h"
#include "GUI/View/Common/CustomEventFilters.h"
#include <QComboBox>
#include <variant>

class QSpinBox;
class QCheckBox;
class UIntDescriptor;
class DoubleSpinBox;
class QDoubleSpinBox;
class DoubleDescriptor;
class QFormLayout;
class ScientificSpinBox;

namespace GUI::Util {

//! Create a combo box with the information found in a ComboProperty.
//!
//! The combo box will be filled with the available options and will get the found tooltips,
//! the common one and individual for each item. The current text will be just passes to "slot"
//! function. Furthermore, the combo box can prohibit accidental changes by the mouse wheel.
//! Otherwise it would be dangerous if the combo is on a scrollable form - unintended and unnoticed
//! changes would take place when just scrolling through the form.
//!
//! The combo can be updated from outside using "updaters" list.
//!
QComboBox* createComboBox(std::function<ComboProperty()> comboFunction,
                          std::function<void(const QString&)> slot,
                          QList<std::function<void()>>* updaters = nullptr, QString tooltip = "",
                          bool isScrollable = true);

//! Create a combo box with the information found in a selection descriptor.
//!
//! The combo will be filled with the available options and will get the found tooltip.
//! The current index will be set according to the current index in the selection.
//! Furthermore, the combo box can prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the combo is on a scrollable form - unintended and unnoticed changes would
//! take place when just scrolling through the form.
//!
//! Changes in the combobox will be notified to the SelectionDescriptor already. The additional (and
//! optional) slot can be used to be notified about an already executed change.
//!
//! The combo can be updated from outside using "updaters" list.
//!
template <typename T>
QComboBox* createComboBoxFromDescriptor(SelectionDescriptor<T> d,
                                        std::function<void(int)> slot = nullptr,
                                        QList<std::function<void()>>* updaters = nullptr,
                                        bool isScrollable = false)
{
    QComboBox* combo = new QComboBox;
    combo->addItems(d.options);
    combo->setMaxCount(d.options.size());
    combo->setToolTip(d.tooltip);
    combo->setCurrentIndex(d.currentIndex());

    if (!isScrollable)
        WheelEventEater::install(combo);

    QObject::connect(combo, qOverload<int>(&QComboBox::currentIndexChanged), [=](int index) {
        d.setCurrentIndex(index);
        if (slot)
            slot(index);
    });

    if (updaters)
        (*updaters) << [=]() {
            QSignalBlocker b(combo);
            combo->setCurrentIndex(d.currentIndex());
        };

    return combo;
}

//! Create a spin box with the information found in a UIntDescriptor.
//!
//! The spin box will be fully initialized (tooltip, limits, current value, size policy).
//! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed changes
//! would take place when just scrolling through the form.
//!
//! No connections to update the descriptor will be established! Therefore changes in the spin box
//! will *not* be notified to the descriptor. The additional (and optional) slot can be used to be
//! notified about a value change.
QSpinBox* createSpinBox(const UIntDescriptor& d, QList<std::function<void()>>* updaters = nullptr,
                        std::function<void(uint)> slot = nullptr, bool easyScrollable = false);

//! Create a label and a spin box with the information found in a UIntDescriptor and place them in a
//! row in a form layout.
//!
//! The label will also contain the unit, if present.
//! Regarding the spin box creation see the method above.
QSpinBox* createSpinBoxRow(QFormLayout* parentLayout, const UIntDescriptor& d);

//! Create a label and a spin box with the information found in a DoubleDescriptor and place them in
//! a row in a form layout.
//!
//! The spin box will be fully initialized (tooltip, limits, current value, unit, size policy).
//! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed changes
//! would take place when just scrolling through the form.
//!
//! No connections to update the descriptor will be established! Therefore changes in the spin box
//! will *not* be notified to the descriptor. The additional (and optional) slot can be used to be
//! notified about a value change (the notified new value is in the base unit of the
//! DoubleDescriptor).
DoubleSpinBox* createDoubleSpinBoxRow(QFormLayout* parentLayout, const DoubleDescriptor& d,
                                      std::function<void(double)> slot = nullptr);

//! Create a label and a scientific spin box with the information found in a DoubleDescriptor and
//! place them in a row in a form layout.
//!
//! The spin box will be fully initialized (tooltip, limits, current value, unit, size policy).
//! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed changes
//! would take place when just scrolling through the form.
//!
//! No connections to update the descriptor will be established! Therefore changes in the spin box
//! will *not* be notified to the descriptor. The additional (and optional) slot can be used to be
//! notified about a value change.
ScientificSpinBox* createScientificSpinBox(QFormLayout* parentLayout, const DoubleDescriptor& d,
                                           std::function<void(double)> slot = nullptr);

//! Create a label with an optional unit in brackets.
//!
//! No trailing ':' will be appended
QString labelWithUnit(const QString& label, std::variant<QString, Unit> unit);

//! Create a label with an optional unit in brackets, both taken from the given descriptor.
//!
//! No trailing ':' will be appended
QString labelWithUnit(const DoubleDescriptor& d);

//! Creates an updatable checkbox
QCheckBox* createCheckBox(const QString& title, std::function<bool()> getter,
                          std::function<void(bool)> setter, QList<std::function<void()>>* updaters);

//! Creates an updatable lineedit
QLineEdit* createTextEdit(std::function<QString()> getter, std::function<void(QString)> setter,
                          QList<std::function<void()>>* updaters);

//! Creates a scrollable updatable spinBox
QSpinBox* createIntSpinbox(std::function<int()> getter, std::function<void(int)> slot,
                           QList<std::function<void()>>* updaters, QString tooltip = "",
                           const RealLimits& limits = RealLimits(), bool easyScrollable = true);

//! Creates a scrollable updatable doublespinBox
QDoubleSpinBox* createDoubleSpinbox(std::function<double()> getter,
                                    std::function<void(double)> slot,
                                    QList<std::function<void()>>* updaters, QString tooltip = "",
                                    const RealLimits& limits = RealLimits(),
                                    bool easyScrollable = true);
} // namespace GUI::Util

#endif // BORNAGAIN_GUI_VIEW_TOOL_WIDGETUTILS_H
