//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Particle/ReCompound.cpp
//! @brief     Implements class ReCompound.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Particle/ReCompound.h"
#include "Base/Util/Algorithms.h"
#include "Base/Vector/WavevectorInfo.h" // debug

ReCompound::~ReCompound()
{
    for (size_t index = 0; index < m_components.size(); ++index)
        delete m_components[index];
}

ReCompound* ReCompound::clone() const
{
    auto* result = new ReCompound();
    for (size_t index = 0; index < m_components.size(); ++index)
        result->addFormFactor(*m_components[index]);
    return result;
}

double ReCompound::radialExtension() const
{
    double result{0.0};
    for (size_t index = 0; index < m_components.size(); ++index)
        result += m_components[index]->radialExtension();
    return result;
}

double ReCompound::bottomZ(const IRotation* rotation) const
{
    if (m_components.empty())
        throw std::runtime_error("ReCompound::bottomZ() -> Error: "
                                 "'this' contains no form factors.");
    return BaseUtils::algo::min_value(
        m_components.begin(), m_components.end(),
        [&rotation](IReParticle* ff) { return ff->bottomZ(rotation); });
}

double ReCompound::topZ(const IRotation* rotation) const
{
    if (m_components.empty())
        throw std::runtime_error("ReCompound::topZ() -> Error: "
                                 "'this' contains no form factors.");
    return BaseUtils::algo::max_value(m_components.begin(), m_components.end(),
                                      [&rotation](IReParticle* ff) { return ff->topZ(rotation); });
}

void ReCompound::addFormFactor(const IReParticle& formfactor)
{
    m_components.push_back(formfactor.clone());
}

void ReCompound::setAmbientMaterial(const Material& material)
{
    for (size_t index = 0; index < m_components.size(); ++index)
        m_components[index]->setAmbientMaterial(material);
}

complex_t ReCompound::theFF(const WavevectorInfo& wavevectors) const
{
    complex_t result(0.0, 0.0);
    for (size_t index = 0; index < m_components.size(); ++index)
        result += m_components[index]->theFF(wavevectors);
    return result;
}

SpinMatrix ReCompound::thePolFF(const WavevectorInfo& wavevectors) const
{
    SpinMatrix result;
    for (size_t index = 0; index < m_components.size(); ++index)
        result += m_components[index]->thePolFF(wavevectors);
    return result;
}
