//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/DepthProbeSimulation.cpp
//! @brief     Implements class DepthProbeSimulation
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/DepthProbeSimulation.h"
#include "Base/Axis/Bin.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Frame.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Beam/IFootprintFactor.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/Distributions.h"
#include "Sim/Computation/DepthProbeComputation.h"

namespace {

const RealLimits alpha_limits = RealLimits::limited(0.0, M_PI_2);
const double zero_phi_i = 0.0;
const double zero_alpha_i = 0.0;

} // namespace


DepthProbeSimulation::DepthProbeSimulation(const MultiLayer& sample)
    : ISimulation(sample)
    , m_beam(Beam::horizontalBeam().clone())
{
}

DepthProbeSimulation::~DepthProbeSimulation() = default;

std::vector<const INode*> DepthProbeSimulation::nodeChildren() const
{
    std::vector<const INode*> result = ISimulation::nodeChildren();
    result.push_back(m_beam.get());
    return result;
}

//! init callbacks for setting the parameter values
void DepthProbeSimulation::initDistributionHandler()
{
    for (const auto& distribution : m_distribution_handler.getDistributions()) {
        const auto* dc = const_cast<ParameterDistribution*>(&distribution);

        switch (dc->whichParameter()) {
        case ParameterDistribution::BeamAzimuthalAngle:
            m_distribution_handler.defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setAzimuthalAngle(d); });
            break;
        case ParameterDistribution::BeamInclinationAngle:
            m_distribution_handler.defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setInclination(d); });
            break;
        case ParameterDistribution::BeamWavelength:
            m_distribution_handler.defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setWavelength(d); });
            break;
        default:
            ASSERT(0);
        }
    }
}

size_t DepthProbeSimulation::numberOfElements() const
{
    return alphaAxis()->size();
}

SimulationResult DepthProbeSimulation::pack_result()
{
    validityCheck();

    auto frame = new Frame({alphaAxis()->clone(), zAxis()->clone()});
    std::vector<double> out(frame->size());
    size_t iout = 0;
    for (size_t i = 0, size = m_depth_eles.size(); i < size; ++i)
        for (const double v : m_depth_eles[i].getIntensities())
            out[iout++] = v;
    auto data = std::make_unique<Datafield>(frame, out);

    return {*data, createCoordSystem()};
}

void DepthProbeSimulation::setZSpan(size_t n_bins, double z_min, double z_max)
{
    if (z_max <= z_min)
        throw std::runtime_error("Error in DepthProbeSimulation::setZSpan: maximum on-axis value "
                                 "is less or equal to the minimum one");
    m_z_axis = std::make_unique<FixedBinAxis>("z", n_bins, z_min, z_max);
}

const IAxis* DepthProbeSimulation::alphaAxis() const
{
    if (!m_alpha_axis)
        throw std::runtime_error("Error in DepthProbeSimulation::alphaAxis: incident angle axis "
                                 "was not initialized.");
    return m_alpha_axis.get();
}

const IAxis* DepthProbeSimulation::zAxis() const
{
    if (!m_z_axis)
        throw std::runtime_error("Error in DepthProbeSimulation::zAxis: position axis "
                                 "was not initialized.");
    return m_z_axis.get();
}

#ifndef SWIG
ICoordSystem* DepthProbeSimulation::createCoordSystem() const
{
    OwningVector<IAxis> axes({m_alpha_axis->clone(), m_z_axis->clone()});
    return new DepthProbeCoordinates(axes, m_beam->direction(), m_beam->wavelength());
}
#endif

void DepthProbeSimulation::setBeamParameters(double lambda, int nbins, double alpha_i_min,
                                             double alpha_i_max, const IFootprintFactor* beam_shape)
{
    FixedBinAxis alpha_axis("alpha_i", static_cast<size_t>(nbins), alpha_i_min, alpha_i_max);

    if (lambda <= 0.0)
        throw std::runtime_error(
            "Error in DepthProbeSimulation::setBeamParameters: wavelength must be positive.");
    if (alpha_axis.min() < 0.0)
        throw std::runtime_error(
            "Error in DepthProbeSimulation::setBeamParameters: minimum value on "
            "angle axis is negative.");
    if (alpha_axis.min() >= alpha_axis.max())
        throw std::runtime_error(
            "Error in DepthProbeSimulation::setBeamParameters: alpha bounds are not sorted.");
    if (alpha_axis.size() == 0)
        throw std::runtime_error(
            "Error in DepthProbeSimulation::setBeamParameters: angle axis is empty");

    m_alpha_axis.reset(alpha_axis.clone());

    // beam is initialized with zero-valued angles
    // Zero-valued incident alpha is required for proper
    // taking into account beam resolution effects
    m_beam->setWavelength(lambda);
    m_beam->setDirection({zero_alpha_i, zero_phi_i});

    if (beam_shape)
        m_beam->setFootprintFactor(*beam_shape);
}

void DepthProbeSimulation::initElementVector()
{
    std::vector<DepthProbeElement> result;

    const double wavelength = m_beam->wavelength();
    const double angle_shift = m_beam->direction().alpha();

    m_depth_eles.clear();
    m_depth_eles.reserve(alphaAxis()->size());
    for (size_t i = 0; i < alphaAxis()->size(); ++i) {
        double result_angle = m_alpha_axis->bin(i).center() + angle_shift;
        m_depth_eles.emplace_back(wavelength, -result_angle, zAxis(),
                                  alpha_limits.isInRange(result_angle));
    }

    if (!m_cache.empty())
        return;
    m_cache.resize(m_depth_eles.size(), std::valarray<double>(0.0, zAxis()->size()));
}

std::unique_ptr<IComputation>
DepthProbeSimulation::createComputation(const reSample& re_sample, size_t start, size_t n_elements)
{
    ASSERT(start < m_depth_eles.size() && start + n_elements <= m_depth_eles.size());
    const auto& begin = m_depth_eles.begin() + static_cast<long>(start);
    return std::make_unique<DepthProbeComputation>(re_sample, options(), progress(), begin,
                                                   begin + static_cast<long>(n_elements));
}

void DepthProbeSimulation::validityCheck() const
{
    const MultiLayer* current_sample = sample();
    if (!current_sample)
        throw std::runtime_error(
            "Error in DepthProbeSimulation::validityCheck: no sample found in the simulation.");

    const size_t data_size = m_depth_eles.size();
    if (data_size != alphaAxis()->size())
        throw std::runtime_error(
            "Error in DepthProbeSimulation::validityCheck: length of simulation "
            "element vector is not equal to the number of inclination angles");
}

void DepthProbeSimulation::validateParametrization(const ParameterDistribution& par_distr) const
{
    const bool zero_mean = par_distr.getDistribution()->mean() == 0.0;
    if (zero_mean)
        return;
    /* TODO come back to this after refactoring of distribution handling
    if (par_distr.whichParameter() == ParameterDistribution::BeamInclinationAngle)
        throw std::runtime_error("Error in DepthProbeSimulation: parameter distribution of "
                                 "beam inclination angle should have zero mean.");
    */
}

void DepthProbeSimulation::normalize(size_t start_ind, size_t n_elements)
{
    const double beam_intensity = m_beam->intensity();
    for (size_t i = start_ind, stop_point = start_ind + n_elements; i < stop_point; ++i) {
        auto& element = m_depth_eles[i];
        const double alpha_i = -element.alphaI();
        const auto* footprint = m_beam->footprintFactor();
        double intensity_factor = beam_intensity;
        if (footprint != nullptr)
            intensity_factor = intensity_factor * footprint->calculate(alpha_i);
        element.setIntensities(element.getIntensities() * intensity_factor);
    }
}

void DepthProbeSimulation::addBackgroundIntensity(size_t, size_t)
{
    if (background())
        throw std::runtime_error(
            "Error: nonzero background is not supported by DepthProbeSimulation");
}

void DepthProbeSimulation::addDataToCache(double weight)
{
    ASSERT(m_depth_eles.size() == m_cache.size());
    for (size_t i = 0, size = m_depth_eles.size(); i < size; ++i)
        m_cache[i] += m_depth_eles[i].getIntensities() * weight;
}

void DepthProbeSimulation::moveDataFromCache()
{
    ASSERT(m_depth_eles.size() == m_cache.size());
    for (size_t i = 0, size = m_depth_eles.size(); i < size; ++i)
        m_depth_eles[i].setIntensities(std::move(m_cache[i]));
    m_cache.clear();
    m_cache.shrink_to_fit();
}
